
"""
All the actions for administering the files and directories in a test suite
"""
from gi.repository import Gtk
import os
import shutil
import subprocess
import re
from texttestlib import plugins, testmodel
from .. import guiplugins, guiutils
from collections import OrderedDict
from zipfile import ZipFile
from fnmatch import fnmatch

# Cut, copy and paste


class FocusDependentAction(guiplugins.ActionGUI):
    def notifyTopWindow(self, window):
        guiplugins.ActionGUI.notifyTopWindow(self, window)
        window.connect("set-focus", self.focusChanged)

    def focusChanged(self, dummy, widget):
        freeTextWidget = isinstance(widget, Gtk.Entry) or isinstance(widget, Gtk.TextView)
        if freeTextWidget:
            self.setSensitivity(False)
        elif self.isActiveOnCurrent():
            self.setSensitivity(True)


class ClipboardAction(FocusDependentAction):
    def isActiveOnCurrent(self, *args):
        if guiplugins.ActionGUI.isActiveOnCurrent(self, *args):
            for test in self.currTestSelection:
                if test.parent:
                    return True
        return False

    def getSignalsSent(self):
        return ["Clipboard"]

    def _getStockId(self):
        return self.getName()

    def _getTitle(self):
        return "_" + self.getName().capitalize()

    def getTooltip(self):
        return self.getName().capitalize() + " selected tests"

    def noAncestorsSelected(self, test):
        if not test.parent:
            return True
        if test.parent in self.currTestSelection:
            return False
        else:
            return self.noAncestorsSelected(test.parent)

    def performOnCurrent(self):
        # If suites are selected, don't also select their contents
        testsForClipboard = list(filter(self.noAncestorsSelected, self.currTestSelection))
        self.notify("Clipboard", testsForClipboard, cut=self.shouldCut())


class CopyTests(ClipboardAction):
    def getName(self):
        return "copy"

    def shouldCut(self):
        return False


class CutTests(ClipboardAction):
    def getName(self):
        return "cut"

    def shouldCut(self):
        return True


class PasteTests(FocusDependentAction):
    def __init__(self, *args):
        FocusDependentAction.__init__(self, *args)
        self.clipboardTests = []
        self.removeAfter = False

    def singleTestOnly(self):
        return True

    def _getStockId(self):
        return "paste"

    def _getTitle(self):
        return "_Paste"

    def getTooltip(self):
        return "Paste tests from clipboard"

    def notifyClipboard(self, tests, cut=False):
        self.clipboardTests = tests
        self.removeAfter = cut
        self.setSensitivity(True)

    def isActiveOnCurrent(self, test=None, state=None):
        return guiplugins.ActionGUI.isActiveOnCurrent(self, test, state) and len(self.clipboardTests) > 0

    def getCurrentTestMatchingApp(self, test):
        # Try the exact application first, if that fails, look for extra-version
        for currTest in self.currTestSelection:
            if currTest.app == test.app:
                return currTest

        for currTest in self.currTestSelection:
            if currTest.app.name == test.app.name:
                return currTest

    def getDestinationInfo(self, test):
        currTest = self.getCurrentTestMatchingApp(test)
        if currTest is None:
            return None, 0
        if currTest.classId() == "test-suite" and currTest not in self.clipboardTests:
            return currTest, 0
        else:
            return currTest.parent, currTest.positionInParent() + 1

    def getNewTestName(self, suite, oldName):
        existingTest = suite.findSubtest(oldName)
        if not existingTest:
            dirName = suite.getNewDirectoryName(oldName)
            if not os.path.exists(dirName):
                return oldName
        elif self.willBeRemoved(existingTest):
            return oldName

        nextNameCandidate = self.findNextNameCandidate(oldName)
        return self.getNewTestName(suite, nextNameCandidate)

    def willBeRemoved(self, test):
        return self.removeAfter and test in self.clipboardTests

    def findNextNameCandidate(self, name):
        copyPos = name.find("_copy_")
        if copyPos != -1:
            copyEndPos = copyPos + 6
            number = int(name[copyEndPos:])
            return name[:copyEndPos] + str(number + 1)
        elif name.endswith("copy"):
            return name + "_2"
        else:
            return name + "_copy"

    def getNewDescription(self, test):
        if len(test.description) or self.removeAfter:
            return test.description
        else:
            return "Copy of " + test.name

    def getRepositionPlacement(self, test, placement):
        currPos = test.positionInParent()
        if placement > currPos:
            return placement - 1
        else:
            return placement

    def messageAfterPerform(self):
        pass  # do it below...

    def performOnCurrent(self):
        newTests = []
        destInfo = OrderedDict()
        for test in self.clipboardTests:
            suite, placement = self.getDestinationInfo(test)
            if suite:
                newName = self.getNewTestName(suite, test.name)
                destInfo[test] = suite, placement, newName
        if len(destInfo) == 0:
            raise plugins.TextTestError(
                "Cannot paste test there, as the copied test and currently selected test have no application/version in common")

        suiteDeltas = {}  # When we insert as we go along, need to update subsequent placements
        for test, (suite, placement, newName) in list(destInfo.items()):
            suiteDeltas.setdefault(suite, 0)
            realPlacement = placement + suiteDeltas.get(suite)
            if self.removeAfter and newName == test.name and suite is test.parent:
                # Cut + paste to the same suite is basically a reposition, do it as one action
                repositionPlacement = self.getRepositionPlacement(test, realPlacement)
                plugins.tryFileChange(test.parent.repositionTest, "Failed to reposition test: no permissions to edit the testsuite file",
                                      test, repositionPlacement)
                newTests.append(test)
            else:
                newDesc = self.getNewDescription(test)
                # Create test files first, so that if it fails due to e.g. full disk, we won't register the test either...
                testDir = suite.getNewDirectoryName(newName)
                try:
                    self.moveOrCopy(test, testDir)
                    suite.registerTest(newName, newDesc, realPlacement)
                    testImported = suite.addTest(test.__class__, os.path.basename(testDir), newDesc, realPlacement)
                    # "testImported" might in fact be a suite: in which case we should read all the new subtests which
                    # might have also been copied
                    testImported.readContents(initial=False)
                    testImported.updateAllRelPaths(test.getRelPath())
                    suiteDeltas[suite] += 1
                    newTests.append(testImported)
                    if self.removeAfter:
                        message = "Failed to remove old test: didn't have sufficient write permission to the test files. Test copied instead of moved."
                        plugins.tryFileChange(test.remove, message)
                except EnvironmentError as e:
                    if os.path.isdir(testDir):
                        shutil.rmtree(testDir)
                    self.showErrorDialog("Failed to paste test:\n" + str(e))

        self.notify("SetTestSelection", newTests)
        self.currTestSelection = newTests
        self.notify("Status", self.getStatusMessage(suiteDeltas))
        if self.removeAfter:
            # After a paste from cut, subsequent pastes should behave like copies of the new tests
            self.notify("Clipboard", newTests, cut=False)
            self.clipboardTests = newTests
            self.removeAfter = False
        for suite, placement, newName in list(destInfo.values()):
            suite.contentChanged()

    def getStatusMessage(self, suiteDeltas):
        suiteName = list(suiteDeltas.keys())[0].name
        if self.removeAfter:
            return "Moved " + self.describeTests() + " to suite '" + suiteName + "'"
        else:
            return "Pasted " + self.describeTests() + " to suite '" + suiteName + "'"

    def getSignalsSent(self):
        return ["SetTestSelection", "Clipboard"]

    def moveOrCopy(self, test, newDirName):
        # If it exists it's because a previous copy has already taken across the directory
        if not os.path.isdir(newDirName):
            oldDirName = test.getDirectory()
            if self.removeAfter:
                self.movePath(oldDirName, newDirName)
            else:
                self.copyPath(oldDirName, newDirName)

    # Methods overridden by version control
    @staticmethod
    def movePath(oldDirName, newDirName):
        os.rename(oldDirName, newDirName)

    @staticmethod
    def copyPath(oldDirName, newDirName):
        shutil.copytree(oldDirName, newDirName, ignore=shutil.ignore_patterns(*plugins.controlDirNames))


# And a generic import test. Note acts on test suites
class ImportTest(guiplugins.ActionDialogGUI):
    def __init__(self, *args):
        guiplugins.ActionDialogGUI.__init__(self, *args)
        self.optionGroup.addOption("name", self.getNameTitle())
        self.optionGroup.addOption("desc", self.getDescTitle(), description="Enter a description of the new " +
                                   self.testType().lower() + " which will be inserted as a comment in the testsuite file.", multilineEntry=True)
        self.optionGroup.addOption("testpos", self.getPlaceTitle(), "last in suite",
                                   allocateNofValues=2, description="Where in the test suite should the test be placed?")
        self.testsImported = []

    def getConfirmationMessage(self):
        testName = self.getNewTestName()
        suite = self.getDestinationSuite()
        self.checkName(suite, testName)
        newDir = os.path.join(suite.getDirectory(), testName)
        if os.path.isdir(newDir):
            if self.testFilesExist(newDir, suite.app):
                raise plugins.TextTestError("Test already exists for application " + suite.app.fullName() +
                                            " : " + os.path.basename(newDir))
            else:
                return "Test directory already exists for '" + testName + "'\nAre you sure you want to use this name?"
        else:
            return ""

    def _getStockId(self):
        return "add"

    def getSizeAsWindowFraction(self):
        # size of the dialog
        return 0.5, 0.45

    def testFilesExist(self, dir, app):
        for fileName in os.listdir(dir):
            parts = fileName.split(".")
            if len(parts) > 1 and parts[1] == app.name:
                return True
        return False

    def singleTestOnly(self):
        return True

    def correctTestClass(self):
        return "test-suite"

    def getNameTitle(self):
        return self.testType() + " Name"

    def getDescTitle(self):
        return self.testType() + " Description"

    def getPlaceTitle(self):
        return "\nPlace " + self.testType()

    def getDefaultName(self):
        return ""

    def getDefaultDesc(self):
        return ""

    def updateOptions(self):
        self.optionGroup.setOptionValue("name", self.getDefaultName())
        self.optionGroup.setOptionValue("desc", self.getDefaultDesc())
        self.setPlacements(self.currTestSelection[0])
        return True

    def setPlacements(self, suite):
        # Add suite and its children
        placements = ["first in suite"]
        for test in suite.testcases:
            placements.append("after " + test.name)
        placements.append("last in suite")

        self.optionGroup.setPossibleValues("testpos", placements)
        self.optionGroup.getOption("testpos").reset()

    def _getTitle(self):
        return "Add " + self.testType()

    def testType(self):  # pragma : no cover - doc only
        return ""

    def messageAfterPerform(self):
        if len(self.testsImported):
            return "Added new " + ", ".join((repr(test) for test in self.testsImported))

    def getNewTestName(self):
        # Overwritten in subclasses - occasionally it can be inferred
        return self.optionGroup.getOptionValue("name").strip()

    def performOnCurrent(self):
        testName = self.getNewTestName()
        description = self.optionGroup.getOptionValue("desc")
        placement = self.getPlacement()
        self.testsImported = []
        for suite in self.currTestSelection:
            suite.registerTest(testName, description, placement)
            testDir = suite.makeSubDirectory(testName)
            self.testsImported.append(self.createTestContents(suite, testDir, description, placement))
            suite.contentChanged()

        self.notify("SetTestSelection", self.testsImported)

    def getSignalsSent(self):
        return ["SetTestSelection"]

    def getDestinationSuite(self):
        return self.currTestSelection[0]

    def getPlacement(self):
        option = self.optionGroup.getOption("testpos")
        return option.possibleValues.index(option.getValue())

    def checkName(self, suite, testName):
        if len(testName) == 0:
            raise plugins.TextTestError("No name given for new " + self.testType() + "!" + "\n" +
                                        "Fill in the 'Adding " + self.testType() + "' tab below.")
        if testName.find(" ") != -1:
            raise plugins.TextTestError("The new " + self.testType() +
                                        " name is not permitted to contain spaces, please specify another")
        for test in suite.testcases:
            if test.name == testName:
                raise plugins.TextTestError("A " + self.testType() + " with the name '" +
                                            testName + "' already exists, please choose another name")


class ImportTestCase(ImportTest):
    def __init__(self, *args):
        ImportTest.__init__(self, *args)
        self.addDefinitionFileOption()

    def testType(self):
        return "Test"

    def addDefinitionFileOption(self):
        self.addOption("opt", "Command line options")

    def createTestContents(self, suite, testDir, description, placement):
        self.writeDefinitionFiles(suite, testDir)
        self.writeEnvironmentFile(suite, testDir)
        self.writeResultsFiles(suite, testDir)
        return suite.addTestCase(os.path.basename(testDir), description, placement)

    def getWriteFileName(self, name, suite, testDir):
        return os.path.join(testDir, name + "." + suite.app.name)

    def getWriteFile(self, name, suite, testDir):
        return open(self.getWriteFileName(name, suite, testDir), "w")

    def writeEnvironmentFile(self, suite, testDir):
        envDir = self.getEnvironment(suite)
        if len(envDir) == 0:
            return
        envFile = self.getWriteFile("environment", suite, testDir)
        for var, value in list(envDir.items()):
            envFile.write(var + ":" + value + "\n")
        envFile.close()

    def writeDefinitionFiles(self, suite, testDir):
        optionString = self.getOptions(suite)
        if len(optionString):
            optionFile = self.getWriteFile("options", suite, testDir)
            optionFile.write(optionString + "\n")
        return optionString

    def getOptions(self, *args):
        return self.optionGroup.getOptionValue("opt")

    def getEnvironment(self, *args):
        return {}

    def writeResultsFiles(self, suite, testDir):
        # Cannot do anything in general
        pass


class ImportTestSuite(ImportTest):
    def __init__(self, *args):
        ImportTest.__init__(self, *args)
        self.addEnvironmentFileOptions()

    def testType(self):
        return "Suite"

    def createTestContents(self, suite, testDir, description, placement):
        return suite.addTestSuite(os.path.basename(testDir), description, placement, self.writeEnvironmentFiles)

    def addEnvironmentFileOptions(self):
        self.addSwitch("env", "Add environment file")

    def writeEnvironmentFiles(self, newSuite):
        if self.optionGroup.getSwitchValue("env"):
            envFile = os.path.join(newSuite.getDirectory(), "environment")
            file = open(envFile, "w")
            file.write("# Dictionary of environment to variables to set in test suite\n")


class ImportApplication(guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        guiplugins.ActionDialogGUI.__init__(self, allApps, dynamic, inputOptions)
        self.fileChooser = None
        self.newApplication = None
        self.inputOptions = inputOptions
        self.rootDirectories = inputOptions.rootDirectories
        self.addOption("name", "Full name of application", description="Name of application to use in reports etc.")
        self.addOption("ext", "\nFile extension to use for TextTest files associated with this application",
                       description="Short space-free extension, to identify all TextTest's files associated with this application")
        possibleSubDirs = self.findSubDirectories()
        self.addOption("subdir", "\nSubdirectory name to store the above application files under (leave blank for local storage)",
                       possibleValues=possibleSubDirs)
        self.addOption("javaclass", "\nJava Class name (instead of executable program)")
        self.addSwitch("gui", "GUI testing option chooser",
                       options=["Disable GUI testing options",
                                "PyGTK GUI with StoryText",
                                "Tkinter GUI with StoryText",
                                "wxPython GUI with StoryText",
                                "SWT GUI with StoryText",
                                "Eclipse RCP GUI with StoryText",
                                "Eclipse GEF GUI with StoryText",
                                "Java Swing GUI with StoryText",
                                "Other embedded Use-case Recorder (e.g. NUseCase)",
                                "Other GUI-test tool (enable virtual display only)"],
                       hideOptions=True)
        self.addSwitch("cpmock", "Enable CaptureMock for record/playback mocking")

        possibleDirs = []
        for app in allApps:
            if app.getDirectory() not in possibleDirs:
                possibleDirs.append(app.getDirectory())
        if len(possibleDirs) == 0:
            possibleDirs = self.rootDirectories
        self.addOption("exec", "\nSelect executable program to test",
                       description="The full path to the program you want to test", possibleDirs=possibleDirs, selectFile=True)

    def createFileChooser(self, *args):
        self.fileChooser = guiplugins.ActionDialogGUI.createFileChooser(self, *args)
        return self.fileChooser

    def onlyDataFilesInFileChooser(self):
        return True

    def createOptionWidget(self, option):
        box, entry = guiplugins.ActionDialogGUI.createOptionWidget(self, option)
        if option is self.optionGroup.getOption("javaclass"):
            entry.connect("changed", self.javaClassChanged)
        return box, entry

    def javaClassChanged(self, *args):
        if self.fileChooser:
            self.setFileChooserSensitivity()

    def setFileChooserSensitivity(self):
        javaclass = self.optionGroup.getValue("javaclass")
        sensitive = self.fileChooser.get_property("sensitive")
        newSensitive = len(javaclass) == 0
        if newSensitive != sensitive:
            self.fileChooser.set_property("sensitive", newSensitive)

    def findSubDirectories(self):
        allDirs = []
        for rootDir in self.rootDirectories:
            usableFiles = [f for f in os.listdir(rootDir) if f not in plugins.controlDirNames]
            allFiles = [os.path.join(rootDir, f) for f in usableFiles]
            allDirs += list(filter(os.path.isdir, allFiles))
        allDirs.sort()
        return list(map(os.path.basename, allDirs))

    def isActiveOnCurrent(self, *args):
        return True

    def _getStockId(self):
        return "add"

    def _getTitle(self):
        return "Add Application"

    def messageAfterPerform(self):
        pass

    def getTooltip(self):
        return "Define a new tested application"

    def checkSanity(self, ext, executable, subdir, directory, javaClass):
        if not ext:
            raise plugins.TextTestError("Must provide a file extension for TextTest files")

        for char in " ./":
            if char in ext:
                raise plugins.TextTestError("File extensions may not contain the character " +
                                            repr(char) + ". It's recommended to stick to alphanumeric characters for this field.")

        if not javaClass and (not executable or not os.path.isfile(executable)):
            raise plugins.TextTestError("Must provide a valid path to a program to test")

        for char in "/\\":
            if char in subdir:
                raise plugins.TextTestError("Subdirectory name must be a local name (not contain " +
                                            repr(char) + ").\nTextTest only looks for applications one level down in the hierarchy.")

        if os.path.exists(os.path.join(directory, "config." + ext)):
            raise plugins.TextTestError(
                "Test-application already exists at the indicated location with the indicated extension: please choose another name")

    def getSignalsSent(self):
        return ["NewApplication"]

    def getMainClass(self, jarFile):
        f = ZipFile(jarFile).open("META-INF/MANIFEST.MF")
        for line in f:
            parts = [part.strip() for part in line.split(":")]
            if len(parts) == 2 and parts[0] == "Main-Class":
                return parts[1]

    def getStatusMessage(self, app):
        return "Created new application '" + app.fullName() + "'."

    def performOnCurrent(self):
        executable = self.optionGroup.getOptionValue("exec")
        ext = self.optionGroup.getOptionValue("ext")
        subdir = self.optionGroup.getOptionValue("subdir")
        directory = self.findFullDirectoryPath(subdir)
        javaClass = self.optionGroup.getOptionValue("javaclass")
        self.checkSanity(ext, executable, subdir, directory, javaClass)
        if executable.startswith(directory):
            executable = executable.replace(directory, "${TEXTTEST_ROOT}")
        plugins.ensureDirectoryExists(directory)
        if javaClass:
            executable = javaClass
        configEntries = OrderedDict({"executable": executable})
        configEntries["filename_convention_scheme"] = "standard"
        if javaClass:
            configEntries["interpreter"] = "java"
        fullName = self.optionGroup.getOptionValue("name")
        if fullName:
            configEntries["full_name"] = fullName
        if self.optionGroup.getSwitchValue("cpmock"):
            configEntries["import_config_file"] = "capturemock_config"
        useGui = self.optionGroup.getSwitchValue("gui")
        if useGui > 0:
            configEntries["use_case_record_mode"] = "GUI"
        if useGui in range(1, 8):
            configEntries["use_case_recorder"] = "storytext"
            storytextDir = os.path.join(directory, "storytext_files")
            plugins.ensureDirectoryExists(storytextDir)
            # Create an empty UI map file so it shows up in the Config tab...
            open(os.path.join(storytextDir, "ui_map.conf"), "w")

            toolkits = ["gtk", "tkinter", "wx", "javaswt", "javarcp", "javagef", "javaswing"]
            toolkit = toolkits[useGui - 1]
            if "java" in toolkit:
                self.setJavaGuiTestingEntries(toolkit, directory, ext, configEntries)
            else:
                self.setPythonGuiTestingEntries(toolkit, directory, ext, configEntries)
        elif useGui == 9:
            configEntries["use_case_recorder"] = "none"

        self.newApplication = self.createApplication(ext, directory, configEntries)
        self.notify("NewApplication", self.newApplication)
        self.notify("Status", self.getStatusMessage(self.newApplication))

    def createApplication(self, ext, directory, configEntries):
        dircache = testmodel.DirectoryCache(directory)
        newApp = testmodel.Application(ext, dircache, [], self.inputOptions, configEntries)
        dircache.refresh()  # we created a config file...
        return newApp

    def respond(self, *args):
        if len(self.validApps) > 0:
            return guiplugins.ActionDialogGUI.respond(self, *args)

    def runDialog(self):
        dialog = self.showConfigurationDialog()
        width = guiutils.guiConfig.getWindowDimension("width", self.diag)
        height = guiutils.guiConfig.getWindowDimension("height", self.diag)
        dialog.resize(width, height)
        while True:
            response = dialog.run()
            if response != Gtk.ResponseType.ACCEPT:
                raise plugins.TextTestError("Application creation cancelled.")

            try:
                self.performOnCurrent()
                break
            except plugins.TextTestError as e:
                self.showErrorDialog(str(e))
        dialog.hide()
        dialog.destroy()
        return self.newApplication, self.getStatusMessage(self.newApplication)

    def getFileOnPath(self, fileName):
        for pathElem in os.getenv("PATH").split(os.pathsep):
            fullPath = os.path.abspath(os.path.join(pathElem, fileName))
            if os.path.isfile(fullPath):
                return fullPath

    def getStoryTextPythonInterpreter(self, toolkit):
        if os.name == "posix":
            return "storytext -i " + toolkit
        else:
            storytextPath = self.getFileOnPath("storytext.py")
            if storytextPath:
                return "python " + storytextPath + " -i " + toolkit
            else:
                raise plugins.TextTestError(
                    "Could not set up Python-GUI testing with StoryText, could not find StoryText installation on PATH")

    def setPythonGuiTestingEntries(self, toolkit, directory, ext, configEntries):
        configEntries["interpreter"] = self.getStoryTextPythonInterpreter(toolkit)
        if toolkit == "gtk":
            comment = "XDG_CONFIG_HOME points to user's ~/.config directory.\n" + \
                      "Behaviour of e.g. FileChoosers can vary wildly depending on settings there.\n" + \
                      "The following settings makes sure it uses an empty test-dependent directory instead."
            configEntries["section_comment"] = comment
            configEntries["copy_test_path"] = "xdg_config_home"
            configEntries["test_data_ignore"] = "xdg_config_home"
            configEntries["test_data_environment"] = [("xdg_config_home", "XDG_CONFIG_HOME")]
        elif toolkit == "tkinter":
            # StoryText doesn't handle tkMessageBox, deal with it via interception by default
            comment = "Tkinter doesn't provide any means to simulate interaction with tkMessageBox.\n" + \
                      "Therefore StoryText cannot handle it. So we capture interaction with it instead.\n" + \
                      "Cannot have multiple threads interacting with tkinter so we disable the threading also."
            configEntries["section_comment"] = comment
            configEntries["import_config_file"] = "capturemock_config"
            cpMockFileName = os.path.join(directory, "capturemockrc." + ext)
            with open(cpMockFileName, "w") as f:
                f.write("[python]\n" +
                        "intercepts = tkMessageBox\n\n" +
                        "[general]\n" +
                        "server_multithreaded = False\n")
        elif toolkit == "wx":
            comment = "wxPython GUIs don't seem to work very well when the hide flag is set on Windows.\n" + \
                      "So we disable it by default here: multiple desktops may be useful."
            configEntries["section_comment"] = comment
            configEntries["virtual_display_hide_windows"] = "false"

    def findStoryTextInPath(self):
        suffix = ".bat" if os.name == "nt" else ""
        for pathElem in os.getenv("PATH").split(os.pathsep):
            storytextPath = os.path.join(pathElem, "storytext")
            jythonPath = os.path.join(pathElem, "jython" + suffix)
            if os.path.isfile(storytextPath) and os.path.isfile(jythonPath):
                return storytextPath, jythonPath
        return None, None

    def findSwingLibraryPath(self, storytextPath):
        rootDir = os.path.dirname(os.path.dirname(storytextPath))
        pattern = "swinglibrary-*.jar"
        for root, _, files in os.walk(rootDir):
            for file in files:
                if fnmatch(file, pattern):
                    return os.path.join(rootDir, root, "*")

    def has_special_chars(self, path):
        return " " in path or "(" in path

    def get_dos_corrected_path(self, path):
        if os.name == "posix" or not self.has_special_chars(path):
            return path

        # Windows paths with spaces don't work in classpaths, figure out the correct DOS path
        dirname, local = os.path.split(path)
        if self.has_special_chars(local):
            p = subprocess.Popen(["dir", "/x"], cwd=dirname, shell=True,
                                 stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
            out = p.communicate()[0]
            for line in out.splitlines():
                if line.endswith(local):
                    parts = line.split()
                    local = parts[3]
                    break
        return os.path.join(self.get_dos_corrected_path(dirname), local)

    def setJavaGuiTestingEntries(self, toolkit, directory, ext, configEntries):
        storytextPath, jythonPath = self.findStoryTextInPath()
        if not storytextPath:
            raise plugins.TextTestError(
                "Could not set up Java-GUI testing with StoryText, could not find StoryText/Jython installation on PATH")
        interpreters = [("jython", jythonPath), ("storytext", storytextPath)]
        configEntries["interpreters"] = interpreters
        # Jython messages, probably don't want to compare them
        configEntries["run_dependent_text"] = [("stderr*", "\\*sys-package-mgr\\*")]
        executable = configEntries["executable"]
        classpath = []
        if executable.endswith(".jar"):
            # Java Guis of one sort or another
            # We don't know how to load jar files directly, store the main class name and add the jar file to the environment.
            mainClass = self.getMainClass(executable)
            if not mainClass:
                raise plugins.TextTestError(
                    "Jar file provided has no main class specified, cannot use it as an executable")
            classpath.append(self.get_dos_corrected_path(executable))
            configEntries["executable"] = mainClass
        if "swing" in toolkit:
            swingLibraryPath = self.findSwingLibraryPath(storytextPath)
            if swingLibraryPath:
                classpath.append(swingLibraryPath)

        if classpath:
            fileName = os.path.join(directory, "environment." + ext)
            with open(fileName, "a") as f:
                f.write("CLASSPATH:" + os.pathsep.join(classpath) + "\n")

        storytextOptionsFile = os.path.join(directory, "storytext_options." + ext)
        with open(storytextOptionsFile, "w") as f:
            f.write("# StoryText options. Run storytext --help for more information on what can be added here\n")
            f.write("-i " + toolkit + "\n")
        with open(storytextOptionsFile + ".debug", "w") as f:
            f.write("-l debug\n")
        jythonOptionsFile = os.path.join(directory, "jython_options." + ext)
        with open(jythonOptionsFile, "w") as f:
            f.write("# Don't use jython's cache\n")
            f.write("-Dpython.cachedir.skip=true\n")
            f.write("# Add Java properties as required, for example:\n")
            f.write("#-Djava.util.prefs.userRoot=preferences\n")
            f.write("# Can also supply JVM arguments by prefixing them with -J, for example:\n")
            f.write("#-J-XX:MaxPermSize=128M")

    def findFullDirectoryPath(self, subdir):
        for rootDir in self.rootDirectories:
            candidate = os.path.normpath(os.path.join(rootDir, subdir))
            if os.path.isdir(candidate):
                return candidate
        return os.path.normpath(os.path.join(self.rootDirectories[0], subdir))

    def getSizeAsWindowFraction(self):
        # size of the dialog
        return 0.7, 0.9


class ImportFiles(guiplugins.ActionDialogGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        self.creationDir = None
        self.defaultAppendAppName = False
        self.appendAppName = False
        self.currentStem = ""
        self.fileChooser = None
        self.newFileInfo = (None, False)
        guiplugins.ActionDialogGUI.__init__(self, allApps, dynamic, inputOptions)
        self.addOption("stem", "Type of file/directory to create", allocateNofValues=2)
        self.addOption("v", "Version identifier to use")
        possibleDirs = self.getPossibleDirs(allApps, inputOptions)
        # The point of this is that it's never sensible as the source for anything, so it serves as a "use the parent" option
        # for back-compatibility
        self.addSwitch("act", options=["Copy file/directory from source",
                                       "Create a new empty file and open it in the editor", "Create a new empty directory"])
        self.addOption("src", "Source to copy from", selectFile=True, possibleDirs=possibleDirs)

    def getPossibleDirs(self, allApps, inputOptions):
        if len(allApps) > 0:
            return sorted(set((app.getDirectory() for app in allApps)))
        else:
            return inputOptions.rootDirectories

    def singleTestOnly(self):
        return True

    def _getTitle(self):
        return "Create/_Import"

    def getTooltip(self):
        return "Create a new file or directory, possibly by copying it"

    def _getStockId(self):
        return "new"

    def getDialogTitle(self):
        return "Create/Import Files and Directories"

    def isActiveOnCurrent(self, *args):
        return self.creationDir is not None and guiplugins.ActionDialogGUI.isActiveOnCurrent(self, *args)

    def getSizeAsWindowFraction(self):
        # size of the dialog
        return 0.7, 0.9

    def getSignalsSent(self):
        return ["NewFile"]

    def messageAfterPerform(self):
        pass

    def updateOptions(self):
        self.currentStem = ""
        return False

    def fillVBox(self, vbox, group):
        test = self.currTestSelection[0]
        dirText = self.getDirectoryText(test)
        self.addText(vbox, "<b><u>" + dirText + "</u></b>")
        self.addText(vbox, "<i>(Test is " + repr(test) + ")</i>")
        return guiplugins.ActionDialogGUI.fillVBox(self, vbox, group)

    def stemChanged(self, *args):
        option = self.optionGroup.getOption("stem")
        newStem = option.getValue()
        if newStem in option.possibleValues and newStem != self.currentStem:
            self.currentStem = newStem
            version = self.optionGroup.getOptionValue("v")
            sourcePath = self.getDefaultSourcePath(newStem, version)
            self.optionGroup.setValue("src", sourcePath)
            if self.defaultAppendAppName:
                self.updateAppendAppName(newStem != "testcustomize.py")

    def actionChanged(self, *args):
        if self.fileChooser:
            self.setFileChooserSensitivity()

    def setFileChooserSensitivity(self):
        action = self.optionGroup.getValue("act")
        sensitive = self.fileChooser.get_property("sensitive")
        newSensitive = action == 0
        if newSensitive != sensitive:
            self.fileChooser.set_property("sensitive", newSensitive)

    def getTargetPath(self, *args, **kwargs):
        targetPathName = self.getFileName(*args, **kwargs)
        return os.path.join(self.creationDir, targetPathName)

    def getDefaultSourcePath(self, stem, version):
        targetPath = self.getTargetPath(stem, version)
        test = self.currTestSelection[0]
        pathNames = test.getAllPathNames(stem, refVersion=version)
        if len(pathNames) > 0:
            firstSource = pathNames[-1]
            if os.path.basename(firstSource).startswith(stem + "." + test.app.name):
                targetPath = self.getTargetPath(stem, version, appendAppName=True)
            if firstSource != targetPath:
                return firstSource
            elif len(pathNames) > 1:
                return pathNames[-2]
        return test.getDirectory()

    def createComboBoxEntry(self, *args):
        combobox, entry = guiplugins.ActionDialogGUI.createComboBoxEntry(self, *args)
        combobox.connect("changed", self.stemChanged)
        return combobox, entry

    def createRadioButtons(self, *args):
        buttons = guiplugins.ActionDialogGUI.createRadioButtons(self, *args)
        buttons[0].connect("toggled", self.actionChanged)
        return buttons

    def createFileChooser(self, *args):
        self.fileChooser = guiplugins.ActionDialogGUI.createFileChooser(self, *args)
        self.fileChooser.set_name("Source File Chooser")
        self.setFileChooserSensitivity()  # Check initial values, maybe set insensitive
        return self.fileChooser

    def addText(self, vbox, text):
        header = Gtk.Label()
        header.set_markup(text + "\n")
        vbox.pack_start(header, False, False, 0)

    def getDirectoryText(self, test):
        relDir = plugins.relpath(self.creationDir, test.getDirectory())
        if relDir:
            return "Create or import files in test subdirectory '" + relDir + "'"
        else:
            return "Create or import files in the test directory"

    def notifyFileCreationInfo(self, creationDir, fileType):
        self.fileChooser = None
        if fileType == "external":
            self.creationDir = None
            self.setSensitivity(False)
        else:
            self.creationDir = creationDir
            newActive = creationDir is not None
            self.setSensitivity(newActive)
            if newActive:
                self.updateStems(fileType)
                self.defaultAppendAppName = (fileType == "definition" or fileType == "approved")
                self.updateAppendAppName(self.defaultAppendAppName)

    def updateAppendAppName(self, setting):
        self.appendAppName = setting
        self.optionGroup.setValue("act", int(setting))

    def findAllStems(self, fileType):
        if fileType == "definition":
            return self.getDefinitionFiles()
        elif fileType == "data":
            return self.currTestSelection[0].getDataFileNames()
        elif fileType == "approved":
            return self.getApprovedFiles()
        else:
            return []

    def getDefinitionFiles(self):
        defFiles = []
        defFiles.append("environment")
        defFiles.append("config")
        defFiles.append("options")
        for interpreter in list(self.currTestSelection[0].getConfigValue("interpreters").keys()):
            defFiles.append(interpreter + "_options")
        if self.currTestSelection[0].classId() == "test-case":
            recordMode = self.currTestSelection[0].getConfigValue("use_case_record_mode")
            if recordMode == "disabled":
                namingScheme = self.currTestSelection[0].getConfigValue("filename_convention_scheme")
                defFiles.append(self.currTestSelection[0].app.getStdinName(namingScheme))
            else:
                defFiles.append("usecase")
        # We only want to create files this way that
        # (a) are not created and understood by TextTest itself ("builtin")
        # (b) are not auto-generated ("regenerate")
        # That leaves the rest ("default")
        return defFiles + self.currTestSelection[0].expandedDefFileStems("default")

    def getApprovedFiles(self):
        app = self.currTestSelection[0].app
        collateKeys = list(app.getConfigValue("collate_file").keys())
        # Don't pick up "dummy" indicators on Windows...
        namingScheme = app.getConfigValue("filename_convention_scheme")
        stdFiles = [app.getStdoutName(namingScheme), app.getStderrName(namingScheme)] + [k for k in collateKeys if k]
        discarded = ["stacktrace"] + app.getConfigValue("discard_file")
        return [f for f in stdFiles if f not in discarded]

    def updateStems(self, fileType):
        stems = self.findAllStems(fileType)
        if len(stems) > 0:
            self.optionGroup.setValue("stem", stems[0])
        else:
            self.optionGroup.setValue("stem", "")
        self.optionGroup.setPossibleValues("stem", stems)

    def getFileName(self, stem, version, appendAppName=False):
        fileName = stem
        if self.appendAppName or appendAppName:
            fileName += "." + self.currTestSelection[0].app.name
        if version:
            fileName += "." + version
        return fileName

    def getConfirmationMessage(self):
        test = self.currTestSelection[0]
        if test.getDirectory() == self.creationDir:
            stemOption = self.optionGroup.getOption("stem")
            stem = stemOption.getValue()
            possible = stemOption.possibleValues
            if possible and stem not in possible:
                if os.sep in stem:
                    firstPart = stem.split(os.sep)[0]  # first component only, can provide a path here
                    if firstPart in possible:
                        return

                return "The type of file/directory you have chosen to create ('" + stem + "') is not currently recognised by TextTest.\n" + \
                    "You can create it anyway, but it may not show up in the GUI and TextTest will probably not use it.\n" + \
                    "Choose one of the drop-down options: " + repr(possible) + ".\n" + \
                    "To fix this, adjust your configuration for 'copy_test_path', 'link_test_path' for data files,\n" + \
                    "'definition_file_stems' for definition files, or 'collate_file' for approved files.\n" + \
                    "Are you sure you want to proceed?"

    def performOnCurrent(self):
        stem = self.optionGroup.getOptionValue("stem")
        version = self.optionGroup.getOptionValue("v")
        action = self.optionGroup.getSwitchValue("act")
        test = self.currTestSelection[0]
        if action > 0:  # Create new
            targetPath = self.getTargetPath(stem, version)
            if os.path.exists(targetPath):
                raise plugins.TextTestError("Not creating file or directory : path already exists:\n" + targetPath)

            if action == 1:
                plugins.ensureDirExistsForFile(targetPath)
                file = open(targetPath, "w")
                file.close()
                test.refreshFiles()
                self.newFileInfo = targetPath, False
            elif action == 2:
                plugins.ensureDirectoryExists(targetPath)
                test.filesChanged()
        else:
            sourcePath = self.optionGroup.getOptionValue("src")
            if sourcePath is None:
                raise plugins.TextTestError(
                    "Must select a valid file or directory to copy, there are none in this empty directory.")
            appendAppName = os.path.basename(sourcePath).startswith(stem + "." + test.app.name)
            targetPath = self.getTargetPath(stem, version, appendAppName)
            if targetPath.startswith(sourcePath) and os.path.isdir(sourcePath):
                raise plugins.TextTestError(
                    "Need to select a valid file or directory to copy, not a parent directory:\n" + sourcePath)
            plugins.ensureDirExistsForFile(targetPath)
            fileExisted = os.path.exists(targetPath)
            plugins.copyPath(sourcePath, targetPath)
            self.newFileInfo = targetPath, fileExisted

    def endPerform(self):
        # Shouldn't start new actions until the current ones complete, framework doesn't like it
        guiplugins.ActionDialogGUI.endPerform(self)
        fileName, existed = self.newFileInfo
        if fileName:
            self.notify("NewFile", fileName, existed)
            self.newFileInfo = (None, False)


class RemoveTests(guiplugins.ActionGUI):
    def __init__(self, *args, **kw):
        self.distinctTestCount = 0
        guiplugins.ActionGUI.__init__(self, *args, **kw)

    def isActiveOnCurrent(self, *args):
        return any((test.parent for test in self.currTestSelection))

    def getActionName(self):
        return "Remove Tests"

    def _getTitle(self):
        return "Remove Tests..."

    def _getStockId(self):
        return "delete"

    def getTooltip(self):
        return "Remove selected tests"

    def getTestCountDescription(self):
        desc = plugins.pluralise(self.distinctTestCount, "test")
        diff = len(self.currTestSelection) - self.distinctTestCount
        if diff > 0:
            desc += " (with " + plugins.pluralise(diff, "extra instance") + ")"
        return desc

    def updateSelection(self, tests, apps, rowCount, *args):
        self.distinctTestCount = rowCount
        return guiplugins.ActionGUI.updateSelection(self, tests, apps, rowCount, *args)

    def getFileRemoveWarning(self):
        return "This will remove files from the file system and hence may not be reversible."

    def getConfirmationMessage(self):
        extraLines = "\n\nNote: " + self.getFileRemoveWarning() + "\n\nAre you sure you wish to proceed?\n"""
        currTest = self.currTestSelection[0]
        if len(self.currTestSelection) == 1:
            if currTest.classId() == "test-case":
                return "\nYou are about to remove the test '" + currTest.name + \
                       "' and all associated files." + extraLines
            else:
                return "\nYou are about to remove the entire test suite '" + currTest.name + \
                       "' and all " + str(currTest.size()) + " tests that it contains." + extraLines
        else:
            return "\nYou are about to remove " + self.getTestCountDescription() + \
                   " and all associated files." + extraLines

    def getTestsToRemove(self, list):
        toRemove = []
        warnings = ""
        for test in list:
            if not test.parent:
                warnings += "\nThe root suite\n'" + test.name + " (" + test.app.name + ")'\ncannot be removed.\n"
                continue
            if test.classId() == "test-suite":
                subTests, subWarnings = self.getTestsToRemove(test.testcases)
                warnings += subWarnings
                for subTest in subTests:
                    if not subTest in toRemove:
                        toRemove.append(subTest)
            if not test in toRemove:
                toRemove.append(test)

        return toRemove, warnings

    def performOnCurrent(self):
        namesRemoved = []
        toRemove, warnings = self.getTestsToRemove(self.currTestSelection)
        permMessage = "Failed to remove test: didn't have sufficient write permission to the test files"
        for test in toRemove:
            dir = test.getDirectory()
            if os.path.isdir(dir):
                plugins.tryFileChange(self.removePath, permMessage, dir)
            if plugins.tryFileChange(test.remove, permMessage):
                namesRemoved.append(test.name)
        self.notify("Status", "Removed test(s) " + ",".join(namesRemoved))
        if warnings:
            self.showWarningDialog(warnings)

    @staticmethod
    def removePath(dir):
        return plugins.removePath(dir)

    def messageAfterPerform(self):
        pass  # do it as part of the method as currentTest will have changed by the end!


class RemoveTestsForPopup(RemoveTests):
    def _getTitle(self):
        return "Remove..."

    def getActionName(self):
        return "Remove Tests For Popup"


class RemoveFiles(guiplugins.ActionGUI):
    def notifyFileCreationInfo(self, creationDir, fileType):
        canRemove = fileType != "external" and \
            (creationDir is None or len(self.currFileSelection) > 0) and \
            self.isActiveOnCurrent()
        self.setSensitivity(canRemove)

    def isActiveOnCurrent(self, *args):
        return len(self.currFileSelection) > 0

    def getActionName(self):
        return "Remove Files"

    def _getTitle(self):
        return "Remove..."

    def _getStockId(self):
        return "delete"

    def getTooltip(self):
        return "Remove selected files"

    def getFileRemoveWarning(self):
        return "This will remove files from the file system and hence may not be reversible."

    def getConfirmationMessage(self):
        extraLines = "\n\nNote: " + self.getFileRemoveWarning() + "\n\nAre you sure you wish to proceed?\n"""
        test = self.getFirstTest()
        return "\nYou are about to remove " + plugins.pluralise(len(self.currFileSelection), self.getType(self.currFileSelection[0][0])) + \
            self.getTestSuffix(test) + "." + extraLines

    def inConfigTab(self):
        return isinstance(self.currFileSelection[0][1], list)

    def getTestSuffix(self, test):
        return " from the " + test.classDescription() + " '" + test.name + "'" if test else ""

    @staticmethod
    def removePath(dir):
        return plugins.removePath(dir)

    def getType(self, filePath):
        if os.path.isdir(filePath):
            return "directory"
        else:
            return "file"

    def getFirstTest(self):
        return self.currTestSelection[0] if not self.inConfigTab() else None

    def performOnCurrent(self):
        test = self.getFirstTest()
        removed = 0
        for filePath, _ in self.currFileSelection:
            fileType = self.getType(filePath)
            self.notify("Status", "Removing " + fileType + " " + os.path.basename(filePath))
            self.notify("ActionProgress")
            permMessage = "Insufficient permissions to remove " + fileType + " '" + filePath + "'"
            if filePath.endswith(".shortcut"):
                from storytext.replayer import ReplayScript
                script = ReplayScript(filePath, False)
                commands = "\n".join(script.commands)
                self.notify("ShortcutRemove", ReplayScript.transformToRegexp(
                    ReplayScript.tryToGetName(filePath)) + "\n renamed to " + commands + "\n")

            if plugins.tryFileChange(self.removePath, permMessage, filePath):
                removed += 1
        if test:
            test.filesChanged()
        else:
            self.appFilesChanged()
        self.notify("Status", "Removed " + plugins.pluralise(removed, fileType) + self.getTestSuffix(test))

    def getSignalsSent(self):
        return ["ReloadConfig", "ShortcutRemove"]

    def appFilesChanged(self):
        appsSeen = set()
        for _, apps in self.currFileSelection:
            for app in apps:
                if app not in appsSeen:
                    appsSeen.add(app)
                    app.refreshFiles()

        self.notify("ReloadConfig")

    def messageAfterPerform(self):
        pass  # do it as part of the method, uses lots of local data


class RepositionTest(guiplugins.ActionGUI):
    def singleTestOnly(self):
        return True

    def _isActiveOnCurrent(self):
        return guiplugins.ActionGUI.isActiveOnCurrent(self) and \
            self.currTestSelection[0].parent and \
            not self.currTestSelection[0].parent.autoSortOrder

    def getSignalsSent(self):
        return ["RefreshTestSelection"]

    def performOnCurrent(self):
        newIndex = self.findNewIndex()
        test = self.currTestSelection[0]
        permMessage = "Failed to reposition test: no permissions to edit the testsuite file"

        if plugins.tryFileChange(test.parent.repositionTest, permMessage, test, newIndex):
            self.notify("RefreshTestSelection")
        else:
            raise plugins.TextTestError("\nThe test\n'" + test.name +
                                        "'\nis not present in the default version\nand hence cannot be reordered.\n")


class RepositionTestDown(RepositionTest):
    def _getStockId(self):
        return "go-down"

    def _getTitle(self):
        return "Move down"

    def messageAfterPerform(self):
        return "Moved " + self.describeTests() + " one step down in suite."

    def getTooltip(self):
        return "Move selected test down in suite"

    def findNewIndex(self):
        return min(self.currTestSelection[0].positionInParent() + 1, self.currTestSelection[0].parent.maxIndex())

    def isActiveOnCurrent(self, *args):
        if not self._isActiveOnCurrent():
            return False
        return self.currTestSelection[0].parent.testcases[self.currTestSelection[0].parent.maxIndex()] != self.currTestSelection[0]


class RepositionTestUp(RepositionTest):
    def _getStockId(self):
        return "go-up"

    def _getTitle(self):
        return "Move up"

    def messageAfterPerform(self):
        return "Moved " + self.describeTests() + " one step up in suite."

    def getTooltip(self):
        return "Move selected test up in suite"

    def findNewIndex(self):
        return max(self.currTestSelection[0].positionInParent() - 1, 0)

    def isActiveOnCurrent(self, *args):
        if not self._isActiveOnCurrent():
            return False
        return self.currTestSelection[0].parent.testcases[0] != self.currTestSelection[0]


class RepositionTestFirst(RepositionTest):
    def _getStockId(self):
        return "goto-top"

    def _getTitle(self):
        return "Move to first"

    def messageAfterPerform(self):
        return "Moved " + self.describeTests() + " to first in suite."

    def getTooltip(self):
        return "Move selected test to first in suite"

    def findNewIndex(self):
        return 0

    def isActiveOnCurrent(self, *args):
        if not self._isActiveOnCurrent():
            return False
        return self.currTestSelection[0].parent.testcases[0] != self.currTestSelection[0]


class RepositionTestLast(RepositionTest):
    def _getStockId(self):
        return "goto-bottom"

    def _getTitle(self):
        return "Move to last"

    def messageAfterPerform(self):
        return "Moved " + repr(self.currTestSelection[0]) + " to last in suite."

    def getTooltip(self):
        return "Move selected test to last in suite"

    def findNewIndex(self):
        return self.currTestSelection[0].parent.maxIndex()

    def isActiveOnCurrent(self, *args):
        if not self._isActiveOnCurrent():
            return False
        currLastTest = self.currTestSelection[0].parent.testcases[len(self.currTestSelection[0].parent.testcases) - 1]
        return currLastTest != self.currTestSelection[0]


class RenameAction(guiplugins.ActionDialogGUI):
    def singleTestOnly(self):
        return True

    def _getStockId(self):
        return "italic"

    def _getTitle(self):
        return "_Rename..."

    def messageAfterPerform(self):
        pass  # Use method below instead.

    def basicNameCheck(self, newName):
        if len(newName) == 0:
            raise plugins.TextTestError("Please enter a new name.")
        if " " in newName:
            raise plugins.TextTestError("The new name must not contain spaces, please choose another name.")

    def performOnCurrent(self):
        try:
            newName = self.optionGroup.getOptionValue("name")
            self.basicNameCheck(newName)
            self.performRename(newName)
        except (IOError, OSError) as e:
            self.showErrorDialog("Failed to " + self.getActionName().lower() + ":\n" + str(e))

    @staticmethod
    def movePath(oldPath, newPath):
        # overridden by version control modules
        os.rename(oldPath, newPath)


class RenameTest(RenameAction):
    def __init__(self, *args):
        RenameAction.__init__(self, *args)
        self.addOption("name", "\nNew name")
        self.addOption("desc", "\nNew description", multilineEntry=True)
        self.oldName = ""
        self.oldDescription = ""

    def getSizeAsWindowFraction(self):
        # size of the dialog
        return 0.5, 0.36

    def isActiveOnCurrent(self, *args):
        # Don't allow renaming of the root suite
        return guiplugins.ActionGUI.isActiveOnCurrent(self, *args) and bool(self.currTestSelection[0].parent)

    def updateOptions(self):
        self.oldName = self.currTestSelection[0].name
        self.oldDescription = self.currTestSelection[0].description
        self.optionGroup.setOptionValue("name", self.oldName)
        self.optionGroup.setOptionValue("desc", self.oldDescription)
        return True

    def fillVBox(self, vbox, group):
        header = Gtk.Label()
        header.set_markup("<b>" + plugins.convertForMarkup(self.oldName) + "</b>")
        vbox.pack_start(header, False, False, 0)
        return guiplugins.ActionDialogGUI.fillVBox(self, vbox, group)

    def getTooltip(self):
        return "Rename selected test"

    def getActionName(self):
        return "Rename Test"

    def getNameChangeMessage(self, newName):
        return "Renamed test " + self.oldName + " to " + newName

    def getChangeMessage(self, newName, newDesc):
        if self.oldName != newName:
            message = self.getNameChangeMessage(newName)
            if self.oldDescription != newDesc:
                message += " and changed description."
            else:
                message += "."
        elif newDesc != self.oldDescription:
            message = "Changed description of test " + self.oldName + "."
        else:
            message = "Nothing changed."
        return message

    def checkNewName(self, newName):
        if newName != self.oldName:
            for test in self.currTestSelection[0].parent.testCaseList():
                if test.name == newName:
                    raise plugins.TextTestError("The name '" + newName +
                                                "' is already taken, please choose another name.")
            newDir = os.path.join(self.currTestSelection[0].parent.getDirectory(), newName)
            if os.path.isdir(newDir):
                self.handleExistingDirectory(newDir)

    def handleExistingDirectory(self, newDir):  # In CVS we might need to override this...
        raise plugins.TextTestError("The directory " + newDir + " already exists, please choose another name.")

    def performRename(self, newName):
        self.checkNewName(newName)
        newDesc = self.optionGroup.getOptionValue("desc")
        if newName != self.oldName or newDesc != self.oldDescription:
            for test in self.currTestSelection:
                # Do this first, so that if we fail we won't update the test suite files either
                self.moveFiles(test, newName)
                test.rename(newName, newDesc)
        changeMessage = self.getChangeMessage(newName, newDesc)
        self.oldName = newName
        self.oldDescription = newDesc
        self.notify("Status", changeMessage)

    def moveFiles(self, test, newName):
        # Create new directory, copy files if the new name is new (we might have
        # changed only the comment ...)
        if test.name != newName:
            oldDir = test.getDirectory()
            newDir = test.parent.getNewDirectoryName(newName)
            if os.path.isdir(oldDir):
                self.movePath(oldDir, newDir)


class RenameFile(RenameAction):
    def __init__(self, *args):
        RenameAction.__init__(self, *args)
        self.addOption("name", "\nNew name for file")
        self.oldName = ""
        self.configAppList = []

    def notifyFileCreationInfo(self, creationDir, fileType):
        canRename = fileType != "external" and \
            (creationDir is None or len(self.currFileSelection) > 0) and \
            self.isActiveOnCurrent()
        self.setSensitivity(canRename)

    def isActiveOnCurrent(self, *args):
        return len(self.currFileSelection) == 1

    def singleTestOnly(self):
        return True

    def updateOptions(self):
        self.oldName = os.path.basename(self.currFileSelection[0][0])
        associatedObj = self.currFileSelection[0][1]
        self.configAppList = associatedObj if isinstance(associatedObj, list) else []
        self.optionGroup.setOptionValue("name", self.oldName)
        return True

    def getSignalsSent(self):
        return ["ReloadConfig", "ShortcutRename"]

    def _getStockId(self):
        return "italic"

    def getActionName(self):
        return "Rename File"

    def _getTitle(self):
        return "_Rename..."

    def getTooltip(self):
        return "Rename selected file"

    def messageAfterPerform(self):
        pass  # Use method below instead.

    def getNameChangeMessage(self, newName):
        return "Renamed file " + self.oldName + " to " + newName + "."

    def checkNewName(self, newName, newPath):
        if newName == self.oldName:
            raise plugins.TextTestError("Please enter a new name.")
        if os.path.exists(newPath):
            raise plugins.TextTestError("There is already a file or directory at '" +
                                        newName + "', please choose another name.")

    def getConfirmationMessage(self):
        oldStem = self.oldName.split(".")[0]
        newName = self.optionGroup.getOptionValue("name")
        newStem = newName.split(".")[0]
        if self.isDefinitionFileStem(oldStem) and not self.isDefinitionFileStem(newStem):
            return "You are trying to rename a definition file in such a way that it will no longer fulfil its previous purpose.\nTextTest uses conventional names for files with certain purposes and '" + oldStem + "' is one such conventional name.\nAre you sure you want to continue?"
        elif self.oldName.endswith(".shortcut"):
            if not newName.endswith(".shortcut"):
                return "A shortcut file name should end with '.shortcut'\nYou are trying to rename it with another extension\nAre you sure you want to continue?"
            elif self.oldName.count("$") != newName.count("$"):
                return "The number of shortcut arguments seems to be different.\nAre you sure you want to continue?"
        else:
            return ""

    def isDefinitionFileStem(self, stem):
        if self.configAppList:
            return stem == "config"
        else:
            return self.currTestSelection[0].isDefinitionFileStem(stem)

    def performRename(self, newName):
        oldPath = self.currFileSelection[0][0]
        newPath = os.path.join(os.path.dirname(oldPath), newName)
        self.checkNewName(newName, newPath)
        self.movePath(oldPath, newPath)
        if self.configAppList:
            for app in self.configAppList:
                app.refreshFiles()
            self.notify("ReloadConfig")
        else:
            self.currTestSelection[0].filesChanged()

        changeMessage = self.getNameChangeMessage(newName)
        if self.oldName.endswith(".shortcut") and newName.endswith(".shortcut"):
            from storytext.replayer import ReplayScript
            self.notify("ShortcutRename", ReplayScript.transformToRegexp(ReplayScript.tryToGetName(
                self.oldName)) + " renamed to " + ReplayScript.tryToGetName(newName))
        self.oldName = newName
        self.notify("Status", changeMessage)


class SortTestSuiteFileAscending(guiplugins.ActionGUI):
    def singleTestOnly(self):
        return True

    def correctTestClass(self):
        return "test-suite"

    def isActiveOnCurrent(self, *args):
        return guiplugins.ActionGUI.isActiveOnCurrent(self, *args) and not self.currTestSelection[0].autoSortOrder

    def _getStockId(self):
        return "sort-ascending"

    def _getTitle(self):
        return "_Sort Test Suite File"

    def messageAfterPerform(self):
        return "Sorted testsuite file for " + self.describeTests() + " in alphabetical order."

    def getTooltip(self):
        return "sort testsuite file for the selected test suite in alphabetical order"

    def performOnCurrent(self):
        self.performRecursively(self.currTestSelection[0], True)

    def performRecursively(self, suite, ascending):
        # First ask all sub-suites to sort themselves
        if guiutils.guiConfig.getValue("sort_test_suites_recursively"):
            for test in suite.testcases:
                if test.classId() == "test-suite":
                    self.performRecursively(test, ascending)

        self.notify("Status", "Sorting " + repr(suite))
        self.notify("ActionProgress")
        if self.hasNonDefaultTests():
            self.showWarningDialog("\nThe test suite\n'" + suite.name +
                                   "'\ncontains tests which are not present in the default version.\nTests which are only present in some versions will not be\nmixed with tests in the default version, which might lead to\nthe suite not looking entirely sorted.")

        suite.sortTests(ascending)

    def hasNonDefaultTests(self):
        if len(self.currTestSelection) == 1:
            return False

        for extraSuite in self.currTestSelection[1:]:
            for test in extraSuite.testcases:
                if not self.currTestSelection[0].findSubtest(test.name):
                    return True
        return False


class SortTestSuiteFileDescending(SortTestSuiteFileAscending):
    def _getStockId(self):
        return "sort-descending"

    def _getTitle(self):
        return "_Reversed Sort Test Suite File"

    def messageAfterPerform(self):
        return "Sorted testsuite file for " + self.describeTests() + " in reversed alphabetical order."

    def getTooltip(self):
        return "sort testsuite file for the selected test suite in reversed alphabetical order"

    def performOnCurrent(self):
        self.performRecursively(self.currTestSelection[0], False)


class ReportBugs(guiplugins.ActionDialogGUI):
    def __init__(self, allApps, *args):
        guiplugins.ActionDialogGUI.__init__(self, allApps, *args)
        self.textGroup = plugins.OptionGroup("Search for")
        self.searchGroup = plugins.OptionGroup("Search in")
        self.applyGroup = plugins.OptionGroup("Additional options to only apply to certain runs")
        self.bugSystemGroup = plugins.OptionGroup("Link failure to a reported bug")
        self.textDescGroup = plugins.OptionGroup("Link failure to a textual description")
        self.textGroup.addOption("search_string", "Text or regexp to match", multilineEntry=True)
        self.textGroup.addSwitch("use_regexp", "Enable regular expressions", 1)
        self.textGroup.addSwitch("trigger_if", name="Trigger if", options=[
                                 "Present", "NOT present", "Exactly as given"])
        self.searchGroup.addSwitch("data_source", options=["Specific file", "Brief text/details", "Full difference report"], description=["Search in a newly generated file (not its diff)",
                                                                                                                                          "Search in the brief text describing the test result as it appears in the Details column in the dynamic GUI test view", "Search in the whole difference report as it appears in the lower right window in the dynamic GUI"])
        self.searchGroup.addOption("search_file", "File to search in")
        self.searchGroup.addSwitch("ignore_other_errors", "Trigger even if other files differ",
                                   description="By default, this bug is only enabled if only the provided file is different. Check this box to enable it irrespective of what other difference exist. Note this increases the chances of it being reported erroneously and should be used carefully.")
        self.searchGroup.addSwitch("trigger_on_success", "Trigger even if file to search would otherwise compare as equal",
                                   description="By default, this bug is only enabled if a difference is detected in the provided file to search. Check this box to search for it even if the file compares as equal.")

        self.applyGroup.addOption("version", "\nVersion to report for")
        self.applyGroup.addOption("execution_hosts", "Trigger only when run on machine(s)")
        self.bugSystemGroup.addOption("bug_system", "\nExtract info from bug system",
                                      "<none>", self.findBugSystems(allApps))
        self.bugSystemGroup.addOption("bug_id", "Bug ID")
        self.textDescGroup.addOption("full_description", "\nFull description")
        self.textDescGroup.addOption("brief_description", "Few-word summary")
        self.textDescGroup.addSwitch("internal_error", "Report as 'internal error' rather than 'known bug'")
        self.optionGroup.addOption(
            "rerun_count", "Number of times to try to rerun the test if the issue is triggered", 0)
        self.optionGroup.addSwitch("rerun_only", "Skip known bug reporting - only use this to trigger reruns")

    def isModal(self):
        return False  # Want to be able to select text from the main GUI while we're in the dialog

    def fillVBox(self, vbox, optionGroup):
        if optionGroup is self.optionGroup:
            for group in [self.textGroup, self.searchGroup, self.applyGroup]:
                if group is self.applyGroup:
                    widget = self.createExpander(group)
                else:
                    widget = self.createFrame(group, group.name)
                vbox.pack_start(widget, False, False, 8)
            vbox.pack_start(Gtk.Separator.new(Gtk.Orientation.HORIZONTAL), True, True, 8)
            header = Gtk.Label()
            header.set_markup("<u>Fill in exactly <i>one</i> of the sections below</u>\n")
            vbox.pack_start(header, False, False, 8)
            for group in [self.bugSystemGroup, self.textDescGroup]:
                frame = self.createFrame(group, group.name)
                vbox.pack_start(frame, False, False, 8)
        return guiplugins.ActionDialogGUI.fillVBox(self, vbox, optionGroup)

    def createExpander(self, group):
        expander = Gtk.Expander.new(group.name)
        expander.add(self.createGroupBox(group))
        return expander

    def createRadioButtons(self, *args):
        buttons = guiplugins.ActionDialogGUI.createRadioButtons(self, *args)
        buttons[0].connect("toggled", self.dataSourceChanged)
        return buttons

    def dataSourceChanged(self, *args):
        sensitive = not self.searchGroup.getOptionValue("data_source")
        self.setGroupSensitivity(self.searchGroup, sensitive)

    def findBugSystems(self, allApps):
        bugSystems = []
        for app in allApps:
            for appSystem in list(app.getConfigValue("bug_system_location").keys()):
                if appSystem not in bugSystems:
                    bugSystems.append(appSystem)
        return bugSystems

    def _getStockId(self):
        return "info"

    def _getTitle(self):
        return "Enter Failure Information"

    def getDialogTitle(self):
        return "Enter information for automatic interpretation of test failures"

    def getDefaultSearchFile(self, possibleValues):
        logFileStems = set([test.getConfigValue("log_file") for test in self.currTestSelection])
        for logFileStem in logFileStems:
            if logFileStem in possibleValues:
                return logFileStem
        return possibleValues[0]

    def _cleanDialog(self, *args):
        self.searchGroup.getOption("search_file").resetDefault()
        guiplugins.ActionDialogGUI._cleanDialog(self, *args)

    def updateOptions(self):
        possibleValues = self.getPossibleFileStems()
        if possibleValues:
            self.searchGroup.setOptionValue("search_file", self.getDefaultSearchFile(possibleValues))

        self.searchGroup.setPossibleValues("search_file", possibleValues)
        return False

    def getPossibleFileStems(self):
        stems = []
        for testOrSuite in self.currTestSelection:
            excludeStems = testOrSuite.expandedDefFileStems()
            for test in testOrSuite.testCaseList():
                for stem in test.dircache.findAllStems():
                    if stem not in stems and stem not in excludeStems:
                        stems.append(stem)
        return stems

    def checkSanity(self):
        searchStr = self.textGroup.getOptionValue("search_string")
        if len(searchStr) == 0:
            raise plugins.TextTestError("Must fill in the field 'text or regexp to match'")
        elif searchStr.startswith(" "):
            raise plugins.TextTestError("'Knownbugs' file format cannot handle leading spaces in search string.\n" +
                                        "If the line starts with spaces, suggest to add a ^ at the start, to match the beginning of the line")
        if not self.optionGroup.getValue("rerun_only"):
            bugSystem = self.bugSystemGroup.getOptionValue("bug_system")
            if self.hasDefaultValue(bugSystem):
                if len(self.textDescGroup.getOptionValue("full_description")) == 0 or \
                   len(self.textDescGroup.getOptionValue("brief_description")) == 0:
                    raise plugins.TextTestError(
                        "Must either provide a bug system or fill in both description and summary fields")
            else:
                if len(self.bugSystemGroup.getOptionValue("bug_id")) == 0:
                    raise plugins.TextTestError("Must provide a bug ID if bug system is given")

    def versionSuffix(self):
        version = self.applyGroup.getOptionValue("version")
        if len(version) == 0:
            return ""
        else:
            return "." + version

    def getFiles(self, ancestors):
        fileNames = []
        for ancestor in ancestors:
            name = "knownbugs." + ancestor.app.name + self.versionSuffix()
            fileName = os.path.join(ancestor.getDirectory(), name)
            if not any((fileName.startswith(f) for f in fileNames)):
                fileNames.append(fileName)
        return [open(fileName, "a") for fileName in fileNames]

    def getSizeAsWindowFraction(self):
        # size of the dialog
        return 0.6, 1.0

    def needsScrollBars(self):
        return True

    @classmethod
    def updateAncestors(cls, ancestors, test):
        for i, ancestor in enumerate(ancestors):
            newAncestor = ancestor.findCommonAncestor(test)
            if newAncestor:
                ancestors[i] = newAncestor
                return True

        return False

    @classmethod
    def findCommonSelectedAncestors(cls, tests):
        ancestors = [tests[0]]
        for test in tests[1:]:
            if not cls.updateAncestors(ancestors, test):
                ancestors.append(test)
        return ancestors

    def performOnCurrent(self):
        self.checkSanity()
        dataSourceText = {1: "brief_text", 2: "free_text"}
        triggerText = {1: "trigger_on_absence", 2: "trigger_on_identical"}
        namesToIgnore = ["version"]
        ancestors = self.findCommonSelectedAncestors(self.currTestSelection)
        for writeFile in self.getFiles(ancestors):
            writeFile.write("\n[Reported by " + os.getenv("USER", "Windows") + " at " + plugins.localtime() + "]\n")
            for group in [self.textGroup, self.searchGroup, self.applyGroup,
                          self.bugSystemGroup, self.textDescGroup, self.optionGroup]:
                for name, option in list(group.options.items()):
                    value = option.getValue()
                    if name in namesToIgnore or self.hasDefaultValue(value, name):
                        continue
                    if name == "data_source":
                        writeFile.write("search_file:" + dataSourceText[value] + "\n")
                        namesToIgnore += ["search_file", "trigger_on_success", "ignore_other_errors"]
                    elif name == "trigger_if":
                        writeFile.write(triggerText[value] + ":" + "1\n")
                    else:
                        writeFile.write(name + ":" + str(value).replace("\n", "\\n") + "\n")
            self.updateWithBugFile(writeFile, ancestors)
            writeFile.close()
        self.setFilesChanged(ancestors)

    def hasDefaultValue(self, value, name=""):
        if name == "use_regexp":
            return value == 1
        else:
            return not value or value in ["0", "<none>"]

    def updateWithBugFile(self, *args):
        pass  # only used in dynamic GUI

    def setFilesChanged(self, ancestors):
        for ancestor in ancestors:
            ancestor.filesChanged()


def getInteractiveActionClasses():
    return [CopyTests, CutTests, PasteTests,
            ImportTestCase, ImportTestSuite, ImportApplication, ImportFiles,
            RenameTest, RenameFile, RemoveTests, RemoveTestsForPopup, RemoveFiles, ReportBugs,
            SortTestSuiteFileAscending, SortTestSuiteFileDescending,
            RepositionTestFirst, RepositionTestUp, RepositionTestDown, RepositionTestLast]




"""
Module to handle the various file-trees in the GUI
"""
from gi.repository import Gtk, GObject
from . import guiutils
import os
import sys
import operator
import logging
import string
from texttestlib import plugins
from collections import OrderedDict
from copy import copy
from functools import reduce


class FileViewGUI(guiutils.SubGUI):
    inheritedText = "(Inherited from parent suites)"

    def __init__(self, dynamic, title="", popupGUI=None):
        guiutils.SubGUI.__init__(self)
        self.model = Gtk.TreeStore(GObject.TYPE_STRING, GObject.TYPE_STRING, GObject.TYPE_STRING,
                                   GObject.TYPE_PYOBJECT, GObject.TYPE_STRING, GObject.TYPE_STRING)
        self.popupGUI = popupGUI
        self.dynamic = dynamic
        self.title = title
        self.selection = None
        self.nameColumn = None
        self.diag = logging.getLogger("File View GUI")

    def recreateModel(self, state, preserveSelection):
        if not self.nameColumn:
            return
        # In theory we could do something clever here, but for now, just wipe and restart
        # Need to re-expand and re-select after clearing...
        if preserveSelection:
            selectionStore = self.storeSelection()
            self.diag.info("Storing " + repr(selectionStore))

        self.model.clear()
        self.addFilesToModel(state)
        view = self.selection.get_tree_view()
        if view:  # Might already be taking down the GUI
            view.expand_all()
        if preserveSelection:
            self.reselect(selectionStore)

    def storeSelection(self):
        selectionStore = []
        self.selection.selected_foreach(self.storeIter, selectionStore)
        return selectionStore

    def storeIter(self, dummyModel, dummyPath, iter, selectionStore):
        selectionStore.append(self._storeIter(iter))

    def _storeIter(self, iter):
        if iter is not None:
            parentStore = self._storeIter(self.model.iter_parent(iter))
            parentStore.append(self.model.get_value(iter, 0))
            return parentStore
        else:
            return []

    def reselect(self, selectionStore):
        for nameList in selectionStore:
            iter = self.findIter(nameList, self.model.get_iter_first())
            if iter is not None:
                self.selection.select_iter(iter)

    def findIter(self, nameList, iter):
        self.diag.info("Looking for iter for " + repr(nameList))
        while iter is not None:
            name = self.model.get_value(iter, 0)
            if name == nameList[0]:
                if len(nameList) == 1:
                    self.diag.info("Succeeded!")
                    return iter
                else:
                    return self.findIter(nameList[1:], self.model.iter_children(iter))
            else:
                iter = self.model.iter_next(iter)
        self.diag.info("Failed!")

    def getState(self):
        pass

    def createView(self):
        self.model.clear()
        state = self.getState()
        self.addFilesToModel(state)
        view = Gtk.TreeView(self.model)
        view.set_name(self.getWidgetName())
        view.set_enable_search(False)  # Shouldn't get big enough to need this
        self.selection = view.get_selection()
        self.selection.set_mode(Gtk.SelectionMode.MULTIPLE)
        self.selection.set_select_function(self.canSelect, self.dynamic)
        renderer = Gtk.CellRendererText()
        self.nameColumn = Gtk.TreeViewColumn(self.title, renderer, text=0, background=1)
        self.nameColumn.set_cell_data_func(renderer, self.renderParentsBold)
        self.nameColumn.set_resizable(True)
        view.append_column(self.nameColumn)
        detailsColumn, recalcRenderer = self.makeDetailsColumn()
        if detailsColumn:
            view.append_column(detailsColumn)
            guiutils.addRefreshTips(view, "file", recalcRenderer, detailsColumn, 5)

        view.expand_all()
        self.monitorEvents()
        view.connect("row_activated", self.fileActivated)

        if self.popupGUI:
            view.connect("button_press_event", self.buttonPressed)
            self.popupGUI.createView()

        view.show()
        return self.addScrollBars(view, hpolicy=Gtk.PolicyType.NEVER)
        # only used in test view

    def buttonPressed(self, *args):
        self.selectionChanged(self.selection)
        self.popupGUI.showMenu(*args)

    def renderParentsBold(self, dummyColumn, cell, model, iter, data):
        if model.iter_has_child(iter):
            cell.set_property('font', "bold")
        else:
            cell.set_property('font', "")

    @staticmethod
    def canSelect(selection, model, path, is_selected, user_data):
        pathIter = model.get_iter(path)
        return not model.iter_has_child(pathIter)

    def makeDetailsColumn(self):
        if self.dynamic:
            renderer = Gtk.CellRendererText()
            column = Gtk.TreeViewColumn("Details")
            column.set_resizable(True)
            recalcRenderer = Gtk.CellRendererPixbuf()
            column.pack_start(renderer, True)
            column.pack_start(recalcRenderer, False)
            column.add_attribute(renderer, 'text', 4)
            column.add_attribute(renderer, 'background', 1)
            column.add_attribute(recalcRenderer, 'stock_id', 5)
            return column, recalcRenderer
        else:
            return None, None

    def fileActivated(self, dummy, path, *args):
        iter = self.model.get_iter(path)
        fileName = self.model.get_value(iter, 2)
        if not fileName:
            # Don't crash on double clicking the header lines...
            return
        comparison = self.model.get_value(iter, 3)
        self.notify(self.getViewFileSignal(), fileName, comparison, False)

    def addFileToModel(self, iter, fileName, colour, associatedObject=None, details=""):
        baseName = os.path.basename(fileName)
        row = [baseName, colour, fileName, associatedObject, details, ""]
        return self.model.insert_before(iter, None, row)

    def addDataFilesUnderIter(self, iter, files, colour, root, **kwargs):
        dirIters = {root: iter}
        parentIter = iter
        for file in files:
            parent = os.path.split(file)[0]
            parentIter = dirIters.get(parent)
            if parentIter is None:
                subDirIters = self.addDataFilesUnderIter(iter, [parent], colour, root)
                parentIter = subDirIters.get(parent)
            newiter = self.addFileToModel(parentIter, file, colour, **kwargs)
            if os.path.isdir(file):
                dirIters[file] = newiter
        return dirIters

    def monitorEvents(self):
        self.selectionChanged(self.selection)
        self.selection.connect("changed", self.selectionChanged)

    def selectionChanged(self, selection):
        filelist = []

        def fileSelected(dummyModel, dummyPath, iter):
            # Do not include the top level which are just headers that don't currently correspond to files
            if self.model.iter_parent(iter) is not None:
                filePath = self.model.get_value(iter, 2)
                if filePath:
                    filelist.append((filePath, self.model.get_value(iter, 3)))

        selection.selected_foreach(fileSelected)
        self.notify("NewFileSelection", filelist)
        if not self.dynamic:
            if selection.count_selected_rows() == 1:
                paths = selection.get_selected_rows()[1]
                selectedIter = self.model.get_iter(paths[0])
                dirName = self.getDirectory(selectedIter)
                fileType = self.getFileType(selectedIter)
                self.notify("FileCreationInfo", dirName, fileType)
            else:
                self.notify("FileCreationInfo", None, None)

    def getFileType(self, iter):
        parent = self.model.iter_parent(iter)
        name = self.model.get_value(iter, 0)
        if name == self.inheritedText:
            return "external"
        elif parent is not None:
            return self.getFileType(parent)
        else:
            return name.split()[0].lower()

    def getDirectory(self, iter):
        fileName = self.model.get_value(iter, 2)
        if fileName:
            if os.path.isdir(fileName):
                return fileName
            else:
                return os.path.dirname(fileName)


class ApplicationFileGUI(FileViewGUI):
    def __init__(self, dynamic, allApps, popupGUI):
        FileViewGUI.__init__(self, dynamic, "Configuration Files", popupGUI)
        self.allApps = copy(allApps)
        self.extras = reduce(operator.add, (app.extras for app in allApps), [])
        self.storytextDirs = {}
        self.testScripts = {}

    def shouldShow(self):
        return not self.dynamic

    def getTabTitle(self):
        return "Config"

    def getViewFileSignal(self):
        return "ViewApplicationFile"

    def getWidgetName(self):
        return "Application File Tree"

    def notifyShortcut(self, *args):
        self.notifyReloadConfig()

    def notifyShortcutRename(self, *args):
        self.notifyReloadConfig()

    def notifyReloadConfig(self):
        for appName, scriptArgs in list(self.testScripts.items()):
            nonexistent = [s for s in scriptArgs if not os.path.isfile(s)]
            for scriptArg in nonexistent:
                scriptArgs.remove(scriptArg)
        self.recreateModel(None, preserveSelection=True)

    def getScriptArgs(self, suite):
        scriptArgs = set()
        scriptCmds = [suite.getConfigValue("executable", expandVars=False)] + \
            list(suite.getConfigValue("interpreters", expandVars=False).values())

        for scriptCmd in scriptCmds:
            for rawScriptArg in scriptCmd.split():
                if "TEXTTEST_ROOT" in rawScriptArg:
                    scriptArg = string.Template(rawScriptArg).safe_substitute(suite.environment)
                    if os.path.isfile(scriptArg):
                        scriptArgs.add(scriptArg)
        return scriptArgs

    def addSuites(self, suites):
        for suite in suites:
            currUsecaseHome = suite.getEnvironment("STORYTEXT_HOME")
            if currUsecaseHome != os.getenv("STORYTEXT_HOME") and os.path.isdir(currUsecaseHome) and currUsecaseHome not in list(self.storytextDirs.values()):
                self.storytextDirs[suite.app] = currUsecaseHome
                if not suite.parent:
                    self.notify("UsecaseHome", suite, currUsecaseHome)

            self.testScripts.setdefault(suite.app.name, set()).update(self.getScriptArgs(suite))
            if suite.app not in self.allApps and suite.app not in self.extras:
                self.allApps.append(suite.app)
                self.recreateModel(self.getState(), preserveSelection=False)

    def getAllFiles(self, storytextDir, app):
        paths = [storytextDir]
        filesToIgnore = app.getCompositeConfigValue("test_data_ignore", os.path.basename(storytextDir))
        for root, dirs, files in os.walk(storytextDir):
            for name in sorted(files) + sorted(dirs):
                if app.fileMatches(name, filesToIgnore):
                    if name in dirs:
                        dirs.remove(name)
                else:
                    paths.append(os.path.join(root, name))
        return paths

    def addFilesToModel(self, *args):
        colour = guiutils.guiConfig.getCompositeValue("file_colours", "static")
        importedFiles = {}
        allTitles = self.getApplicationTitles()
        for index, app in enumerate(self.allApps):
            headerRow = ["Files for " + allTitles[index], "white", app.getDirectory(), None, "", ""]
            confiter = self.model.insert_before(None, None, headerRow)
            for file in self.getConfigFiles(app):
                self.addFileToModel(confiter, file, colour, [app])
                for importedFile in self.getImportedFiles(file, app):
                    importedFiles[importedFile] = importedFile
            storytextDir = self.storytextDirs.get(app)
            if storytextDir:
                files = self.getAllFiles(storytextDir, app)
                self.addDataFilesUnderIter(confiter, files, colour,
                                           app.getDirectory(), associatedObject=self.allApps)
            testScripts = self.testScripts.get(app.name)
            if testScripts:
                headerRow = ["Scripts for " + allTitles[index], "white", app.getDirectory(), None, "", ""]
                scriptiter = self.model.insert_before(None, None, headerRow)
                self.addDataFilesUnderIter(scriptiter, sorted(testScripts), colour,
                                           app.getDirectory(), associatedObject=self.allApps)

        # Handle recursive imports here ...

        if len(importedFiles) > 0:
            importediter = self.model.insert_before(None, None)
            self.model.set_value(importediter, 0, "Imported Files")
            sortedFiles = list(importedFiles.values())
            sortedFiles.sort()
            for importedFile in sortedFiles:
                self.addFileToModel(importediter, importedFile, colour, self.allApps)

        personalDir = plugins.getPersonalConfigDir()
        personalFiles = self.getPersonalFiles(personalDir)
        if len(personalFiles) > 0:
            headerRow = ["Personal Files", "white", personalDir, None, "", ""]
            persiter = self.model.insert_before(None, None, headerRow)
            self.addDataFilesUnderIter(persiter, personalFiles, colour, personalDir, associatedObject=self.allApps)

    def getApplicationTitles(self):
        basicTitles = [repr(app) for app in self.allApps]
        if self.areUnique(basicTitles):
            return basicTitles
        else:
            return [repr(app) + " (" + app.name + " under " +
                    os.path.basename(app.getDirectory()) + ")" for app in self.allApps]

    def areUnique(self, names):
        for index, name in enumerate(names):
            for otherName in names[index + 1:]:
                if name == otherName:
                    return False
        return True

    def getConfigFiles(self, app):
        dircaches = app.getAllDirCaches("config", [app.dircache])
        return app.getAllFileNames(dircaches, "config", allVersions=True)

    def getPersonalFiles(self, personalDir):
        if not os.path.isdir(personalDir):
            return []
        allFiles = []
        for root, dirs, files in os.walk(personalDir):
            for pruneDir in ["tmp"] + plugins.controlDirNames:
                if pruneDir in dirs:
                    dirs.remove(pruneDir)
            for file in files + dirs:
                allFiles.append(os.path.join(root, file))
        return sorted(allFiles)

    def getImportedFiles(self, file, app=None):
        imports = []
        if os.path.isfile(file):
            importLines = [l for l in open(file, "r").readlines() if l.startswith("import_config_file")]
            for line in importLines:
                try:
                    file = line.split(":")[1].strip()
                    if app:
                        file = app.configPath(file)
                    if os.path.isfile(file):
                        imports.append(file)
                except Exception:  # App. file not found ...
                    continue
        return imports


class TestFileGUI(FileViewGUI):
    def __init__(self, dynamic, popupGUI):
        FileViewGUI.__init__(self, dynamic, "", popupGUI)
        self.currentTest = None

    @staticmethod
    def canSelect(selection, model, path, is_selected, user_data):
        if user_data:
            pathIter = model.get_iter(path)
            return model.iter_parent(pathIter) is not None
        else:
            return True

    def getTabTitle(self):
        return "Test"

    def getWidgetName(self):
        return "File Tree"

    def getViewFileSignal(self):
        return "ViewFile"

    def notifyNameChange(self, test, origRelPath):
        if test is not self.currentTest:
            return

        def updatePath(model, dummyPath, iter):
            origFile = model.get_value(iter, 2)
            if origFile:
                newFile = origFile.replace(origRelPath, test.getRelPath())
                model.set_value(iter, 2, newFile)

        self.model.foreach(updatePath)
        self.setName([test], 1)

    def notifyFileChange(self, test):
        if test is self.currentTest:
            self.recreateModel(test.stateInGui, preserveSelection=True)

    def notifyLifecycleChange(self, test, state, changeDesc):
        if test is self.currentTest:
            self.recreateModel(state, preserveSelection=changeDesc.find("approve") == -1)

    def notifyRecalculation(self, test, comparisons, newIcon):
        if test is not self.currentTest:
            return

        def setRecalculateIcon(model, dummyPath, iter):
            comparison = model.get_value(iter, 3)
            if comparison in comparisons:
                oldVal = model.get_value(iter, 5)
                if oldVal != newIcon:
                    self.model.set_value(iter, 5, newIcon)

        self.model.foreach(setRecalculateIcon)

    def forceVisible(self, rowCount):
        return rowCount == 1

    def notifyNewTestSelection(self, tests, dummyApps, rowCount, *args):
        if len(tests) == 0 or (not self.dynamic and rowCount > 1):  # multiple tests in static GUI result in removal
            self.currentTest = None
            self.setName(tests, rowCount)
            self.model.clear()
            return

        if len(tests) > 1 and self.currentTest in tests:
            self.setName(tests, rowCount)
        else:
            self.currentTest = tests[0]
            self.currentTest.refreshFiles()
            self.setName(tests, rowCount)
            # New test selected, keep file selection!
            # See TTT-2273. Previously we didn't keep the file selection, unclear why...
            self.recreateModel(self.getState(), preserveSelection=True)

    def notifySetFileSelection(self, fileStems):
        def trySelect(model, dummyPath, iter):
            comparison = model.get_value(iter, 3)
            if comparison is not None and comparison.stem in fileStems:
                self.selection.select_iter(iter)
            else:
                self.selection.unselect_iter(iter)

        self.model.foreach(trySelect)

    def notifyViewerStarted(self):
        self.applicationEvent("the viewer process to start", timeDelay=1)

    def notifyNewFile(self, fileName, overwrittenExisting):
        if os.path.isfile(fileName):
            self.notify(self.getViewFileSignal(), fileName, None, True)
        if not overwrittenExisting or os.path.isdir(fileName):
            self.currentTest.refreshFiles()
            self.recreateModel(self.getState(), preserveSelection=True)

    def setName(self, tests, rowCount):
        newTitle = self.getName(tests, rowCount)
        if newTitle != self.title:
            self.title = newTitle
            if self.nameColumn:
                self.nameColumn.set_title(self.title)

    def getName(self, tests, rowCount):
        if rowCount > 1:
            return "Sample from " + repr(len(tests)) + " tests"
        elif self.currentTest:
            return self.currentTest.name.replace("_", "__")
        else:
            return "No test selected"

    def getColour(self, name):
        return guiutils.guiConfig.getCompositeValue("file_colours", name)

    def shouldShowCurrent(self, *args):
        return self.currentTest is not None

    def addFilesToModel(self, state):
        if not state:
            return
        realState = state
        if state.isMarked():
            realState = state.oldState
        if self.dynamic:
            if hasattr(realState, "correctResults"):
                # failed on comparison
                self.addComparisonsToModel(realState)
            else:
                self.addTmpFilesToModel(realState)
        else:
            self.addStaticFilesToModel(realState)

    def getState(self):
        if self.currentTest:
            return self.currentTest.stateInGui

    def addComparisonsToModel(self, state):
        self.addComparisons(state, state.correctResults + state.changedResults, "Comparison Files")
        self.addComparisons(state, state.newResults, "New Files")
        self.addComparisons(state, state.missingResults, "Missing Files")

    def addComparisons(self, state, compList, title):
        if len(compList) == 0:
            return
        iter = self.model.insert_before(None, None)
        self.model.set_value(iter, 0, title)
        filelist = []
        splitlist = {}
        fileCompMap = {}
        for comp in compList:
            file = comp.getDisplayFileName()
            fileCompMap[file] = comp
            if comp.getParent():
                splitlist.setdefault(comp.getParent().getDisplayFileName(), []).append(comp)
            else:
                filelist.append(file)

        for file in sorted(filelist):
            newiter = self.addApprovedFileUnderIter(state, iter, file, fileCompMap)
            splitcomps = splitlist.get(file, [])
            # Preserve the original order of these split files
            for splitcomp in sorted(splitcomps, key=state.allResults.index):
                self.addApprovedFileUnderIter(state, newiter, splitcomp.getDisplayFileName(), fileCompMap)

    def addApprovedFileUnderIter(self, state, iter, file, compMap={}):
        comparison = compMap.get(file)
        colour = self.getComparisonColour(state, comparison)
        details = ""
        if comparison:
            details = comparison.getDetails()
        return self.addFileToModel(iter, file, colour, comparison, details)

    def getComparisonColour(self, state, fileComp):
        if not state.isComplete():
            return self.getColour("running")
        if fileComp and fileComp.hasSucceeded():
            return self.getColour("success")
        else:
            return self.getColour("failure")

    def addTmpFilesToModel(self, state):
        tmpFiles = self.currentTest.listTmpFiles()
        tmpIter = self.model.insert_before(None, None)
        self.model.set_value(tmpIter, 0, "Temporary Files")
        for file in tmpFiles:
            self.addApprovedFileUnderIter(state, tmpIter, file)

    def getRootIterAndColour(self, heading, rootDir=None):
        if not rootDir:
            rootDir = self.currentTest.getDirectory()
        headerRow = [heading + " Files", "white", rootDir, None, "", ""]
        stditer = self.model.insert_before(None, None, headerRow)
        colour = guiutils.guiConfig.getCompositeValue("file_colours", "static_" + heading.lower(), defaultKey="static")
        return stditer, colour

    def addStaticFilesWithHeading(self, heading, stdFiles):
        stditer, colour = self.getRootIterAndColour(heading)
        for file in stdFiles:
            self.addFileToModel(stditer, file, colour)

    def addStaticFilesToModel(self, *args):
        stdFiles, defFiles = self.currentTest.listApprovedFiles(allVersions=True)
        if self.currentTest.classId() == "test-case":
            self.addStaticFilesWithHeading("Approved", stdFiles)

        self.addStaticFilesWithHeading("Definition", defFiles)
        self.addStaticDataFilesToModel()
        self.addExternallyEditedFilesToModel()
        self.addExternalFilesToModel()

    def getExternalDataFiles(self):
        try:
            return list(self.currentTest.app.extraReadFiles(self.currentTest).items())
        except Exception:
            sys.stderr.write("WARNING - ignoring exception thrown by '" + self.currentTest.getConfigValue("config_module") +
                             "' configuration while requesting extra data files, not displaying any such files\n")
            return OrderedDict()

    def addStaticDataFilesToModel(self):
        if len(self.currentTest.getDataFileNames()) == 0:
            return
        datiter, colour = self.getRootIterAndColour("Data")
        dataFiles, inheritedDataFiles = self.currentTest.listDataFilesWithInherited()
        currDir = self.currentTest.getDirectory()
        self.addDataFilesUnderIter(datiter, dataFiles, colour, currDir)
        if inheritedDataFiles:
            inheritedRow = [self.inheritedText, "white", currDir, None, "", ""]
            inheritedIter = self.model.insert_before(datiter, None, inheritedRow)
            for suite, inherited in list(inheritedDataFiles.items()):
                self.addDataFilesUnderIter(inheritedIter, inherited, colour, suite.getDirectory())

    def addExternalFilesToModel(self):
        externalFiles = self.getExternalDataFiles()
        if len(externalFiles) == 0:
            return
        datiter, colour = self.getRootIterAndColour("External")
        for name, filelist in externalFiles:
            exiter = self.model.insert_before(datiter, None)
            self.model.set_value(exiter, 0, name)
            self.model.set_value(exiter, 1, "white")  # mostly to trigger output...
            for file in filelist:
                self.addFileToModel(exiter, file, colour)

    def addExternallyEditedFilesToModel(self):
        root, files = self.currentTest.listExternallyEditedFiles()
        if root:
            datiter, colour = self.getRootIterAndColour("Externally Edited", root)
            self.addDataFilesUnderIter(datiter, files, colour, root)



#!/usr/bin/env python3

import optparse
import os
import sys
import io
import difflib


def fixSysPath(fileName):
    install_root = os.path.dirname(os.path.dirname(fileName))
    # We pick up the basic libraries.
    # or a "generic" directory containing the TextTest core with local modules in the root
    for subdir in ["", "lib", "site/lib", "generic"]:
        libDir = os.path.abspath(os.path.join(install_root, subdir))
        if os.path.isdir(libDir):
            sys.path.insert(0, libDir)


def main():
    fixSysPath(os.path.abspath(__file__))
    from texttestlib.default.fpdiff import fpfilter
    parser = optparse.OptionParser("usage: %prog [options] fromfile tofile")
    parser.add_option("-t", "--tolerance", type="float",
                      help='Set absolute floating point tolerance')
    parser.add_option("-r", "--relative", type="float",
                      help='Set relative floating point tolerance')
    parser.add_option("-o", "--output",
                      help='Write filtered tofile to use external diff')
    parser.add_option("-d", "--difflib", action="store_true", default=False,
                      help="Use python's difflib")
    parser.add_option("-s", "--split", default='',
                      help='Compare individual fields after splitting at the given separator, use "None" to emulate split()')
    (options, args) = parser.parse_args()
    if len(args) == 0:  # pragma: no cover - not production code
        parser.print_help()
        sys.exit(1)
    if len(args) != 2:  # pragma: no cover - not production code
        parser.error("need to specify both a fromfile and tofile")
    fromfile, tofile = args
    fromlines = open(fromfile).readlines()
    tolines = open(tofile).readlines()
    if options.output:
        out = open(options.output, 'w')
        fpfilter(fromlines, tolines, out, options.tolerance, options.relative, options.difflib, options.split)
        out.close()
    else:  # pragma: no cover - not production code
        out = io.StringIO()
        fpfilter(fromlines, tolines, out, options.tolerance, options.relative, options.difflib, options.split)
        out.seek(0)
        tolines = out.readlines()
        sys.stdout.writelines(difflib.unified_diff(fromlines, tolines, fromfile, tofile))


if __name__ == '__main__':
    main()




"""
The various text info views, i.e. the bottom right-corner "Text Info" and
the "Run Info" tab from the dynamic GUI
"""
from gi.repository import Gtk, GObject, Gdk, Pango
import os
import sys
import datetime
from . import guiutils, guiplugins
from texttestlib import plugins
from texttestlib.default import performance


class TimeMonitor:
    def __init__(self):
        self.timingInfo = {}

    def notifyLifecycleChange(self, test, dummyState, changeDesc):
        if changeDesc in ["start", "complete"]:
            self.timingInfo.setdefault(test, []).append((changeDesc, datetime.datetime.now()))

    def shouldShow(self):
        # Nothing to show, but needed to be a GUI observer
        return True

    def getElapsedTime(self, test):
        timingInfo = self.timingInfo.get(test)
        if timingInfo:
            delta = datetime.datetime.now() - timingInfo[0][1]
            return delta.seconds + delta.days * 60 * 60 * 24
        else:
            return -1

    def getTimingReport(self, test):
        timingInfo = self.timingInfo.get(test)
        text = ""
        if timingInfo:
            text += "\n"
            for desc, timestamp in timingInfo:
                descToUse = self.getTimeDescription(desc)
                text += descToUse + ": " + timestamp.strftime(plugins.datetimeFormat) + "\n"
        return text

    def getTimeDescription(self, changeDesc):
        descToUse = changeDesc.replace("complete", "end").capitalize() + " time"
        return descToUse.ljust(17)


class TextViewGUI(guiutils.SubGUI):
    hovering_over_link = False
    hand_cursor = Gdk.Cursor.new(Gdk.CursorType.HAND2)
    regular_cursor = Gdk.Cursor.new(Gdk.CursorType.XTERM)
    linkMarker = "URL=http"
    timeMonitor = TimeMonitor()

    def __init__(self, dynamic):
        guiutils.SubGUI.__init__(self)
        self.dynamic = dynamic
        self.text = ""
        self.showingSubText = False
        self.view = None

    def shouldShowCurrent(self, *args):
        return len(self.text) > 0

    def forceVisible(self, rowCount):
        # Both TextInfo and RunInfo should stay visible when tests are selected
        return rowCount == 1

    def updateView(self):
        if self.view:
            self.updateViewFromText(self.text)

    def createView(self):
        self.view = Gtk.TextView()
        self.view.set_name(self.getTabTitle())
        self.view.set_editable(False)
        self.view.set_cursor_visible(False)
        self.view.set_wrap_mode(Gtk.WrapMode.WORD)
        self.updateViewFromText(self.text)
        self.view.show()
        return self.addScrollBars(self.view, hpolicy=Gtk.PolicyType.AUTOMATIC)

    def updateViewFromText(self, text):
        textbuffer = self.view.get_buffer()
        if self.linkMarker in text:
            self.view.connect("event-after", self.event_after)
            self.view.connect("motion-notify-event", self.motion_notify_event)
            self.setHyperlinkText(textbuffer, text)
        else:
            textbuffer.set_text(text)

    def getEnvironmentLookup(self):
        pass

    # Links can be activated by clicking. Low-level code lifted from Maik Hertha's
    # GTK hypertext demo
    def event_after(self, text_view, event):  # pragma : no cover - external code and untested browser code
        if event.type != Gdk.EventType.BUTTON_RELEASE:
            return False
        if event.button != 1:
            return False
        buffer = text_view.get_buffer()

        # we shouldn't follow a link if the user has selected something
        try:
            start, end = buffer.get_selection_bounds()
        except ValueError:
            # If there is nothing selected, None is return
            pass
        else:
            if start.get_offset() != end.get_offset():
                return False

        x, y = text_view.window_to_buffer_coords(Gtk.TextWindowType.WIDGET, int(event.x), int(event.y))
        iter = text_view.get_iter_at_location(x, y)
        target = self.findLinkTarget(iter)
        if target:
            statusMessage = guiplugins.openLinkInBrowser(target)
            self.notify("Status", statusMessage)

        return False

    # Looks at all tags covering the position (x, y) in the text view,
    # and if one of them is a link, change the cursor to the "hands" cursor
    # typically used by web browsers.
    def set_cursor_if_appropriate(self, text_view, x, y):  # pragma : no cover - external code
        hovering = False

        iter = text_view.get_iter_at_location(x, y)

        hovering = bool(self.findLinkTarget(iter))
        if hovering != self.hovering_over_link:
            self.hovering_over_link = hovering

        if self.hovering_over_link:
            text_view.get_window(Gtk.TextWindowType.TEXT).set_cursor(self.hand_cursor)
        else:
            text_view.get_window(Gtk.TextWindowType.TEXT).set_cursor(self.regular_cursor)

    def findLinkTarget(self, iter):  # pragma : no cover - called by external code
        tags = iter.get_tags()
        for tag in tags:
            target = tag.get_data("target")
            if target:
                return target

    # Update the cursor image if the pointer moved.
    def motion_notify_event(self, text_view, event):  # pragma : no cover - external code
        x, y = text_view.window_to_buffer_coords(Gtk.TextWindowType.WIDGET,
                                                 int(event.x), int(event.y))
        self.set_cursor_if_appropriate(text_view, x, y)
        text_view.window.get_pointer()
        return False

    def setHyperlinkText(self, buffer, text):
        buffer.set_text("", 0)
        iter = buffer.get_iter_at_offset(0)
        for line in text.splitlines():
            if self.linkMarker in line:
                self.insertLinkLine(buffer, iter, line)
            else:
                buffer.insert(iter, line + "\n")

    def insertLinkLine(self, buffer, iter, line):
        # Assumes text description followed by link
        tag = buffer.create_tag(None, foreground="blue", underline=Pango.Underline.SINGLE)
        words = line.strip().split()
        linkTarget = words[-1][4:]  # strip off the URL=
        newLine = " ".join(words[:-1]) + "\n"
        tag.target = linkTarget
        buffer.insert_with_tags(iter, newLine, tag)

    def getDescriptionText(self, test):
        paragraphs = self.getDescriptionParagraphs(test)
        return "\n\n".join(paragraphs)

    def getDescriptionParagraphs(self, test):
        paragraphs = [self.getDescription(test)]
        for stem in sorted(set(["performance"] + list(test.getConfigValue("performance_logfile_extractor").keys()))):
            fileName = test.getFileName(stem)
            if fileName and os.path.isfile(fileName):
                paragraphs.append(self.getFilePreview(fileName))
        return paragraphs

    def getFilePreview(self, fileName):
        return "Expected " + os.path.basename(fileName).split(".")[0] + " for the default version:\n" + \
               performance.describePerformance(fileName)

    def getDescription(self, test):
        header = "Description:\n"
        if test.description:
            return header + test.description
        else:
            return header + "<No description provided>"


class RunInfoGUI(TextViewGUI):
    def __init__(self, dynamic, runName, reconnect):
        TextViewGUI.__init__(self, dynamic)
        self.reconnect = reconnect
        self.text = "Information will be available here when all tests have been read..."
        self.runName = runName

    def getTabTitle(self):
        return "Run Info"

    def shouldShow(self):
        return self.dynamic

    def appInfo(self, suite):
        textToUse = "Application name : " + suite.app.fullName() + "\n"
        textToUse += "Version          : " + suite.app.getFullVersion() + "\n"
        textToUse += "Number of tests  : " + str(suite.size()) + "\n"
        if not self.reconnect:
            textToUse += "Executable       : " + suite.getConfigValue("executable") + "\n"
        return textToUse

    def notifySetRunName(self, name):
        self.runName = name
        if self.view:
            self.updateView()

    def updateView(self):
        if self.runName:
            self.updateViewFromText("Run Name         : " + self.runName + "\n\n" + self.text)
        else:
            self.updateViewFromText(self.text)

    def notifyAllRead(self, suites):
        self.text = ""
        self.text += "\n".join(map(self.appInfo, suites)) + "\n"
        self.text += "Command line     : " + plugins.commandLineString(sys.argv) + "\n\n"
        self.text += "Start time       : " + plugins.startTimeString() + "\n"
        self.updateView()

    def notifyAllComplete(self):
        self.text += "End time         : " + plugins.localtime() + "\n"
        self.updateView()


class TestRunInfoGUI(TextViewGUI):
    def __init__(self, dynamic, reconnect):
        TextViewGUI.__init__(self, dynamic)
        self.currentTest = None
        self.reconnect = reconnect
        self.resetText()

    def shouldShow(self):
        return self.dynamic and not self.reconnect

    def getTabTitle(self):
        return "Test Run Info"

    def notifyNewTestSelection(self, tests, *args):
        if len(tests) == 0:
            self.currentTest = None
            self.resetText()
        elif self.currentTest not in tests:
            self.currentTest = tests[0]
            self.resetText()

    def resetText(self):
        self.text = "Selected test  : "
        if self.currentTest:
            self.text += self.currentTest.getRelPath() + "\n"
            self.appendTestInfo(self.currentTest)
        else:
            self.text += "none\n"
        self.updateView()

    def appendTestInfo(self, test):
        self.text += self.timeMonitor.getTimingReport(test)
        self.text += "\n" + self.getDescriptionText(test)
        if test.classId() == "test-case":
            self.text += "\n\n" + test.app.getRunDescription(test)


class TextInfoGUI(TextViewGUI):
    def __init__(self, *args):
        TextViewGUI.__init__(self, *args)
        self.currentTest = None
        self.currFileSelection = []
        self.preambleText = ""

    def getTabTitle(self):
        return "Text Info"

    def resetText(self, test, state):
        if state.category == "not_started":
            self.text = "\n" + self.getDescriptionText(self.currentTest)
            self.preambleText = self.text
        else:
            self.text = ""
            freeText = state.getFreeText()
            if state.isComplete():
                self.text = "Test " + repr(state) + "\n"
                if len(freeText) == 0:
                    self.text = self.text.replace(" :", "")
                self.preambleText = self.text
            self.text += str(freeText)
            if state.hasStarted() and not state.isComplete():
                self.text += self.getPerformanceEstimate(test)
                self.text += "\n\nTo obtain the latest progress information and an up-to-date comparison of the files above, " + \
                             "perform 'recompute status' (press '" + \
                             guiutils.guiConfig.getCompositeValue("gui_accelerators", "recompute_status") + "')"

    def getPerformanceEstimate(self, test):
        expected = performance.getTestPerformance(test)
        if expected > 0:
            elapsed = self.timeMonitor.getElapsedTime(test)
            if elapsed >= 0:
                perc = (elapsed * 100) / expected
                return "\nReckoned to be " + str(int(perc)) + "% complete comparing elapsed time with expected performance.\n" + \
                       "(" + performance.getTimeDescription(elapsed) + \
                    " of " + performance.getTimeDescription(expected) + ")"
        return ""

    def getDescriptionParagraphs(self, test):
        paragraphs = TextViewGUI.getDescriptionParagraphs(self, test)
        testPath = test.getRelPath()
        if testPath:  # Don't include this for root suite
            paragraphs.insert(1, "Full path:\n" + testPath)
        return paragraphs

    def notifyNewTestSelection(self, tests, *args):
        if len(tests) == 0:
            self.currentTest = None
            self.preambleText = ""
            self.text = "No test currently selected"
            self.updateView()
        elif self.currentTest not in tests:
            self.currentTest = tests[0]
            self.resetText(self.currentTest, self.currentTest.stateInGui)
            self.updateView()

    def notifyDescriptionChange(self, *args):
        self.resetText(self.currentTest, self.currentTest.stateInGui)
        self.notifyNewFileSelection(self.currFileSelection)
        self.updateView()

    def notifyLifecycleChange(self, test, state, *args):
        if not test is self.currentTest:
            return
        self.resetText(test, state)
        self.updateView()

    def notifyRefreshFilePreviews(self, test, fileName):
        if self.isSelected(fileName):
            self.notifyNewFileSelection(self.currFileSelection)

    def isSelected(self, fileName):
        return any((currFile == fileName for currFile, _ in self.currFileSelection))

    def makeSubText(self, files):
        newText = self.preambleText
        showComparisonPreviews = self.dynamic and not all((comp and comp.hasSucceeded() for _, comp in files))
        for fileName, comp in files:
            if showComparisonPreviews:
                if comp and not comp.hasSucceeded():
                    newText += comp.getFreeText()
            elif os.path.isfile(fileName):
                newText += self.getPreview(fileName)
        return newText, newText != self.text and newText != self.preambleText

    def getPreview(self, fileName):
        baseName = os.path.basename(fileName)
        stem = baseName.split(".")[0]
        text = "\n\nPreview of " + baseName + ":\n"
        if self.currentTest.configValueMatches("binary_file", stem):
            text += "Contents of file are marked as binary via the 'binary_file' setting."
        else:
            maxLength = self.currentTest.getConfigValue("lines_of_text_difference")
            maxWidth = self.currentTest.getConfigValue("max_width_text_difference")
            previewGenerator = plugins.PreviewGenerator(maxWidth, maxLength)
            text += previewGenerator.getPreview(open(fileName))
        return text

    def notifyNameChange(self, test, *args):
        if test is self.currentTest:
            self.resetText(self.currentTest, self.currentTest.stateInGui)
            self.notifyNewFileSelection(self.currFileSelection)
            self.updateView()

    def notifyNewFileSelection(self, files):
        self.currFileSelection = files
        if len(files) == 0:
            if self.showingSubText:
                self.showingSubText = False
                self.updateViewFromText(self.text)
        elif self.preambleText:
            newText, changed = self.makeSubText(files)
            if changed:
                self.showingSubText = True
                self.updateViewFromText(newText)
            elif self.showingSubText:
                self.showingSubText = False
                self.updateViewFromText(self.text)

    def getEnvironmentLookup(self):
        if self.currentTest:
            return self.currentTest.environment.get



import datetime
import time
import os
from texttestlib import plugins


class BatchVersionFilter:
    def __init__(self, batchSession):
        self.batchSession = batchSession

    def verifyVersions(self, app):
        badVersion = self.findUnacceptableVersion(app)
        if badVersion is not None:
            raise plugins.TextTestWarning("unregistered version '" + badVersion +
                                          "' for " + self.batchSession + " session.")

    def findUnacceptableVersion(self, app):
        if app.getCompositeConfigValue("batch_use_version_filtering", self.batchSession) != "true":
            return

        allowedVersions = app.getCompositeConfigValue("batch_version", self.batchSession)
        for version in app.versions:
            if len(version) > 0 and version not in allowedVersions and not version.startswith("copy_"):
                return version


def getBatchRunName(optionMap):
    name = optionMap.get("name")
    if name is not None:
        return name

    jenkinsBuildNumber = os.getenv("BUILD_NUMBER")
    timeToUse = plugins.globalStartTime
    if jenkinsBuildNumber is None:
        # If we're not using Jenkins, assume some kind of nightjob set up (mostly for historical reasons)
        # Here we use a standardised date that give a consistent answer for night-jobs.
        # Hence midnight is a bad cutover point. The day therefore starts and ends at 8am :)
        timeToUse -= datetime.timedelta(hours=8)
    name = timeToUse.strftime("%d%b%Y")
    if jenkinsBuildNumber is not None:
        name += "." + jenkinsBuildNumber
    return name


def parseFileName(fileName, diag):
    versionStr = fileName[5:-5]
    components = versionStr.split("_")
    diag.info("Parsing file with components " + repr(components))
    for index, component in enumerate(components[1:]):
        try:
            diag.info("Trying to parse " + component + " as date")
            date = time.strptime(component, "%d%b%Y")
            version = "_".join(components[:index + 1])
            tag = "_".join(components[index + 2:]) or component
            return version, date, tag
        except ValueError:
            pass
    return None, None, None


def getEnvironmentFromRunFiles(runNameDirs, tag):
    env = {}
    for dir in runNameDirs:
        path = os.path.join(dir, tag)
        if os.path.isfile(path):
            with open(path) as f:
                for line in f:
                    if "=" in line:
                        var, val = line.strip().split("=", 1)
                        env[var] = val
    return env


def convertToUrl(path, fileMapping):
    for filePath, httpPath in list(fileMapping.items()):
        if path.startswith(filePath):
            return path.replace(filePath, httpPath)
    return "file://" + os.path.abspath(path)




"""
The actions in the dynamic GUI that affect the state of a test
"""
from gi.repository import Gtk
import os
from texttestlib import plugins
from .. import guiplugins
from .adminactions import ReportBugs
from texttestlib.default.knownbugs import CheckForBugs, BugMap
from configparser import ConfigParser
from copy import copy
from glob import glob
from threading import Thread


class BackgroundThreadHelper:
    backgroundThread = None

    def getSignalsSent(self):
        return ["BackgroundActionCompleted"]

    def messageAfterPerform(self):
        pass

    def performInBackground(self):
        selection = copy(self.currTestSelection)
        if self.backgroundThread and self.backgroundThread.is_alive():
            self.notify("Status", "Waiting for previous background action to finish ...")
            self.backgroundThread.join()
        BackgroundThreadHelper.backgroundThread = Thread(target=self.runBackgroundThread, args=(selection,))
        self.backgroundThread.start()

    def runBackgroundThread(self, *args):
        self.notify("ActionStart", False)
        try:
            errorMsg = self.performBackgroundAction(*args)
        except plugins.TextTestError as e:
            errorMsg = str(e)
        self.notify("ActionStop", False)
        self.notify("BackgroundActionCompleted", errorMsg, self)

    def _runInteractive(self):
        guiplugins.ActionGUI._runInteractive(self)
        self.performInBackground()

    def notifyBackgroundActionCompleted(self, message, origAction):
        if message and origAction is self:
            self.showErrorDialog(message)


class ApproveTests(BackgroundThreadHelper, guiplugins.ActionDialogGUI):
    defaultVersionStr = "<existing version>"
    fullVersionStr = "<full version>"

    def __init__(self, allApps, *args):
        guiplugins.ActionDialogGUI.__init__(self, allApps, *args)
        self.directAccel = None
        self.directAction = Gtk.Action("Approve", "_Approve",
                                       self.getDirectTooltip(), self.getStockId())
        self.directAction.connect("activate", self._respond)
        self.directAction.set_property("sensitive", False)
        self.addOption("v", "Version to approve", self.defaultVersionStr)
        self.addOption("old", "Version(s) to store previous results as")
        self.addSwitch("over", "Replace successfully compared files also", 0)
        if self.hasPerformance(allApps):
            self.addSwitch("ex", "Store", 1, ["Average performance", "Exact performance"])
        # Must do this in the constructor, so that "Approve" also takes account of them
        for option in list(self.optionGroup.options.values()):
            self.addValuesFromConfig(option)

    def createDialog(self):
        dialog = guiplugins.ActionDialogGUI.createDialog(self)
        dialog.set_name("Approve As")
        return dialog

    def getDialogTitle(self):
        stemsToApprove = self.getStemsToApprove()
        saveDesc = "Approving " + str(len(self.currTestSelection)) + " tests"
        if len(stemsToApprove) > 0:
            saveDesc += ", only files " + ",".join(stemsToApprove)
        return saveDesc

    def _getStockId(self):
        return "apply"

    def _getTitle(self):
        return "Approve _As..."

    def getTooltip(self):
        return "Approve results with non-default settings"

    def getDirectTooltip(self):
        return "Approve results for selected tests. This will save the files produced by this run as the new approved files " + \
            "which future runs will be compared with. Used to be called Save."""

    def addToGroups(self, actionGroup, accelGroup):
        self.directAccel = self._addToGroups("Approve", self.directAction, actionGroup, accelGroup)
        guiplugins.ActionDialogGUI.addToGroups(self, actionGroup, accelGroup)

    def setSensitivity(self, newValue):
        self._setSensitivity(self.directAction, newValue)
        self._setSensitivity(self.gtkAction, newValue)
        if newValue:
            self.updateOptions()

    def getConfirmationMessage(self):
        testsForWarn = [test for test in self.currTestSelection if test.stateInGui.warnOnSave()]
        if len(testsForWarn) == 0:
            return ""
        message = "You have selected tests whose results are partial or which are registered as bugs:\n"
        for test in testsForWarn:
            message += "  Test '" + test.uniqueName + "' " + test.stateInGui.categoryRepr() + "\n"
        message += "Are you sure you want to do this?\n"
        return message

    def getSaveableTests(self, selection):
        return [test for test in selection if test.stateInGui.isSaveable()]

    def updateOptions(self):
        versionOption = self.optionGroup.getOption("v")
        currOption = versionOption.defaultValue
        newVersions = self.getPossibleVersions()
        currVersions = versionOption.possibleValues
        if self.defaultVersionStr == currOption and newVersions == currVersions:
            return False
        self.optionGroup.setOptionValue("v", versionOption.defaultValue)
        self.diag.info("Setting default save version to " + self.defaultVersionStr)
        self.optionGroup.setPossibleValues("v", newVersions)
        return True

    def getPossibleVersions(self):
        extensions = [self.defaultVersionStr, self.fullVersionStr]
        for app in self.currAppSelection:
            for ext in app.getSaveableVersions():
                if not ext in extensions:
                    extensions.append(ext)
        # Include the default version always
        extensions.append("")
        return extensions

    def getExactness(self):
        return int(self.optionGroup.getSwitchValue("ex", 1))

    def isActiveOnCurrent(self, test=None, state=None):
        if state and state.isSaveable():
            return True
        for seltest in self.currTestSelection:
            if seltest is not test and seltest.stateInGui.isSaveable():
                return True
        return False

    def getStemsToApprove(self):
        return [cmp.stem for _, cmp in self.currFileSelection]

    def getBackupVersions(self):
        versionString = self.optionGroup.getOptionValue("old")
        if versionString:
            return plugins.commasplit(versionString)
        else:
            return []

    def getVersion(self, test):
        versionString = self.optionGroup.getOptionValue("v")
        if versionString == self.fullVersionStr:
            return test.app.getFullVersion(forSave=1)
        elif versionString != self.defaultVersionStr:
            return versionString

    def performOnCurrent(self):
        backupVersions = self.getBackupVersions()
        if self.optionGroup.getOptionValue("v") in backupVersions:
            raise plugins.TextTestError(
                "Cannot backup to the same version we're trying to approve! Choose another name.")

    def performBackgroundAction(self, selection):
        backupVersions = self.getBackupVersions()
        stemsToApprove = self.getStemsToApprove()
        overwriteSuccess = self.optionGroup.getSwitchValue("over")
        tests = self.getSaveableTests(selection)
        # Calculate the versions beforehand, as approving tests can change the selection,
        # which can affect the default version calculation...
        testsWithVersions = [(test, self.getVersion(test)) for test in tests]
        testDesc = str(len(tests)) + " tests"
        self.notify("Status", "Approving " + testDesc + " ...")
        try:
            for test, versionString in testsWithVersions:
                testComparison = test.stateInGui
                testComparison.setObservers(self.observers)
                testComparison.save(test, self.getExactness(), versionString,
                                    overwriteSuccess, stemsToApprove, backupVersions)
                newState = testComparison.makeNewState(test, "approved")
                test.changeState(newState)

            self.notify("Status", "Approved " + testDesc + ".")
        except OSError as e:
            self.notify("Status", "Failed to approve " + testDesc + ".")
            errorStr = str(e)
            if "Permission" in errorStr:
                errorStr = "Failed to approve " + testDesc + \
                    " : didn't have sufficient write permission to the test files"
            return errorStr


class SplitResultFiles(guiplugins.ActionGUI):
    def __init__(self, *args):
        guiplugins.ActionGUI.__init__(self, *args)
        self.latestTestCount = 0

    def isActiveOnCurrent(self, test=None, state=None):
        for currTest in self.currTestSelection:
            separators = currTest.getConfigValue("file_split_pattern")
            if separators:
                if currTest is test:
                    if state.isComplete():
                        return True
                elif currTest.stateInGui.isComplete():
                    return True
        return False

    def _getTitle(self):
        return "Split result files"

    def _getStockId(self):
        return "convert"

    def getTooltip(self):
        return "Split result files to be able to handle different changes separately"

    def messageAfterPerform(self):
        if self.latestTestCount == 0:
            return "No test was split."
        else:
            return "Split result files for " + plugins.pluralise(self.latestTestCount, "test") + "."

    def performOnCurrent(self):
        self.latestTestCount = 0
        for test in self.currTestSelection:
            separators = test.getConfigValue("file_split_pattern")
            if separators:
                self.latestTestCount += 1
                self.notify("Status", "Splitting result files for " + repr(test) + " ...")
                self.notify("ActionProgress")
                if test.stateInGui.hasResults():
                    self.splitResultFiles(test, test.stateInGui, separators)

    def splitResultFiles(self, test, state, separators):
        newComparisons = state.splitResultFiles(test, separators)
        if newComparisons:
            newState = state.makeNewState(test, "recalculated")
            for comp in newComparisons:
                newState.addComparison(comp)
            test.changeState(newState)


class RecomputeTests(BackgroundThreadHelper, guiplugins.ActionGUI):
    def isActiveOnCurrent(self, test=None, state=None):
        for currTest in self.currTestSelection:
            if currTest is test:
                if state.hasStarted():
                    return True
            elif currTest.stateInGui.hasStarted():
                return True
        return False

    def _getTitle(self):
        return "Recompute Status"

    def _getStockId(self):
        return "refresh"

    def getTooltip(self):
        return "Recompute test status, including progress information if appropriate"

    def getSignalsSent(self):
        return ["Recomputed"] + BackgroundThreadHelper.getSignalsSent(self)

    def performOnCurrent(self):
        if any((test.stateInGui.isComplete() for test in self.currTestSelection)):
            self.reloadConfigForSelected()

    def performBackgroundAction(self, selection):
        latestTestCount = 0
        for test in selection:
            latestTestCount += 1
            self.notify("Status", "Recomputing status of " + repr(test) + " ...")
            test.app.recomputeProgress(test, test.stateInGui, self.observers)
            self.notify("Recomputed", test)

        self.notify("Status", self.getFinalMessage(latestTestCount))

    def getFinalMessage(self, latestTestCount):
        if latestTestCount == 0:
            return "No test needed recomputation."
        else:
            return "Recomputed status of " + plugins.pluralise(latestTestCount, "test") + "."


class MarkTest(guiplugins.ActionDialogGUI):
    def __init__(self, *args):
        guiplugins.ActionDialogGUI.__init__(self, *args)
        self.addOption("brief", "Brief text", "Checked")
        self.addOption("free", "Free text", "Checked at " + plugins.localtime())

    def _getTitle(self):
        return "_Mark"

    def getTooltip(self):
        return "Mark the selected tests"

    def performOnCurrent(self):
        for test in self.currTestSelection:
            self.notifyMark(test, self.optionGroup.getOptionValue("free"), self.optionGroup.getOptionValue("brief"))

    def notifyMark(self, test, freeText, briefText):
        oldState = test.stateInGui
        if oldState.isComplete():
            if test.stateInGui.isMarked():
                oldState = test.stateInGui.oldState  # Keep the old state so as not to build hierarchies ...
            newState = plugins.MarkedTestState(freeText, briefText, oldState)
            test.changeState(newState)
            self.notify("ActionProgress")  # Just to update gui ...

    def isActiveOnCurrent(self, test=None, state=None):
        if state and state.isComplete():
            return True
        for seltest in self.currTestSelection:
            if seltest is not test and seltest.stateInGui.isComplete():
                return True
        return False


class UnmarkTest(guiplugins.ActionGUI):
    def _getTitle(self):
        return "_Unmark"

    def getTooltip(self):
        return "Unmark the selected tests"

    def performOnCurrent(self):
        for test in self.currTestSelection:
            if test.stateInGui.isMarked():
                test.stateInGui.oldState.lifecycleChange = "unmarked"  # To avoid triggering completion ...
                test.changeState(test.stateInGui.oldState)
                self.notify("ActionProgress")  # Just to update gui ...

    def isActiveOnCurrent(self, *args):
        for test in self.currTestSelection:
            if test.stateInGui.isMarked():
                return True
        return False


class LoadFromRerun(guiplugins.ActionGUI):
    def _getTitle(self):
        return "Load from Rerun"

    def getRerunMarkedTests(self):
        return [test for test in self.currTestSelection if test.stateInGui.isMarked() and test.stateInGui.briefText.startswith("Rerun")]

    def performOnCurrent(self):
        for test in self.getRerunMarkedTests():
            rerunNumber = test.stateInGui.briefText.split()[-1]
            pattern = os.path.join(os.getenv("TEXTTEST_TMP"), "*." + rerunNumber + "_from_" +
                                   plugins.startTimeString().replace(":", "") + "*")
            dirs = glob(pattern)
            if len(dirs) == 1:
                rerunDir = dirs[0]
                appDir = os.path.join(rerunDir, test.app.name + test.app.versionSuffix())
                if not os.path.isdir(appDir):
                    allAppDirs = glob(os.path.join(rerunDir, test.app.name + "*"))
                    if len(allAppDirs) > 0:
                        allAppVersions = [d.replace(os.path.join(rerunDir, test.app.name + "."), "")
                                          for d in allAppDirs]
                        raise plugins.TextTestError("Cannot load data for rerun, version '" + test.app.getFullVersion() +
                                                    "' could not be found, only version(s) '" + ", ".join(allAppVersions) + "'")
                if self.loadFrom(test, appDir):
                    continue
            raise plugins.TextTestError("Cannot load data for rerun, test " + test.name +
                                        " has not yet completed or has been deleted in rerun " + rerunNumber)

    def loadFrom(self, test, appDir):
        from texttestlib.default.reconnect import ReconnectTest
        reconn = ReconnectTest(appDir, False)
        location = os.path.join(appDir, test.getRelPath())
        state = reconn.getReconnectStateFrom(test, location, copyEvenIfLoadFails=False)
        if state:
            state.lifecycleChange = "recalculated"
            test.backupTemporaryData()
            reconn.copyFiles(test, location)
            test.changeState(state)
            return True
        else:
            return False

    def notifyRerunDirectory(self, app, rerunNumber, directory):
        self.rerunDirectories[(app, rerunNumber)] = directory

    def isActiveOnCurrent(self, *args):
        return len(self.getRerunMarkedTests()) > 0


class SuspendTests(guiplugins.ActionGUI):
    def _getTitle(self):
        return "Suspend"

    def getTooltip(self):
        return "Suspend the selected tests"

    def performOnCurrent(self):
        from texttestlib.queuesystem.masterprocess import QueueSystemServer
        QueueSystemServer.instance.setSuspendStateForTests(self.currTestSelection, True)

    def isActiveOnCurrent(self, *args):
        return any((not test.stateInGui.isComplete() for test in self.currTestSelection))


class UnsuspendTests(guiplugins.ActionGUI):
    def _getTitle(self):
        return "Unsuspend"

    def getTooltip(self):
        return "Unsuspend the selected tests"

    def performOnCurrent(self):
        from texttestlib.queuesystem.masterprocess import QueueSystemServer
        QueueSystemServer.instance.setSuspendStateForTests(self.currTestSelection, False)

    def isActiveOnCurrent(self, *args):
        return any((not test.stateInGui.isComplete() for test in self.currTestSelection))


class KillTests(guiplugins.ActionGUI):
    def _getStockId(self):
        return "stop"

    def _getTitle(self):
        return "_Kill"

    def getTooltip(self):
        return "Kill selected tests"

    def isActiveOnCurrent(self, test=None, state=None):
        for seltest in self.currTestSelection:
            if seltest is test:
                if not state.isComplete():
                    return True
            else:
                if not seltest.stateInGui.isComplete():
                    return True
        return False

    def getSignalsSent(self):
        return ["Kill"]

    def performOnCurrent(self):
        tests = [test for test in self.currTestSelection if not test.stateInGui.isComplete()]
        tests.reverse()  # best to cut across the action thread rather than follow it and disturb it excessively
        testDesc = str(len(tests)) + " tests"
        self.notify("Status", "Killing " + testDesc + " ...")
        for test in tests:
            self.notify("ActionProgress")
            test.notify("Kill")

        self.notify("Status", "Killed " + testDesc + ".")


def applyBugsToTests(tests, bugMap):
    foundMatch = False
    for test in tests:
        newState, rerunCount = CheckForBugs().checkTestWithBugs(test, test.stateInGui, bugMap)
        if newState:
            newState.lifecycleChange = "recalculated"
            test.changeState(newState)
        if newState or rerunCount:
            foundMatch = True

    return foundMatch


class ReportBugsAndRecompute(ReportBugs):
    def getFiles(self, *args):
        from io import StringIO
        return [StringIO()]

    def updateOptions(self):
        ReportBugs.updateOptions(self)
        multifile = int(len(self.getPossibleFileStems()) > 1)
        self.searchGroup.setOptionValue("ignore_other_errors", multifile)
        if not self.currTestSelection[0].stateInGui.hasResults():
            self.searchGroup.setSwitchValue("data_source", 2)

    def fillVBox(self, *args):
        ret = ReportBugs.fillVBox(self, *args)
        self.dataSourceChanged()
        return ret

    def getPossibleFileStems(self):
        state = self.currTestSelection[0].stateInGui
        if state.hasResults():
            return [comp.stem for comp in state.allResults if not comp.hasSucceeded()]
        else:
            return []

    def updateWithBugFile(self, bugFile, ancestors):
        bugFile.seek(0)
        bugMap = BugMap()
        bugMap.readFromFileObject(bugFile)
        if applyBugsToTests(self.currTestSelection, bugMap):
            for realFile in ReportBugs.getFiles(self, ancestors):
                realFile.write(bugFile.getvalue())
                realFile.close()
        else:
            raise plugins.TextTestError("Information entered did not trigger on the selected test, please try again")


class FindKnownBugs(guiplugins.ActionDialogGUI):
    def __init__(self, allApps, *args):
        guiplugins.ActionDialogGUI.__init__(self, allApps, *args)
        self.optionGroup.addOption("bug", "Bug or Brief Text", allocateNofValues=2)
        self.optionGroup.addSwitch("copy",
                                   options=["Apply to whole suite",
                                            "Apply to common parent suite", "Copy info into test(s)"],
                                   description=["Move the reported failure information to the root suite where it will apply to all tests",
                                                "Move the reported failure information to the suite where it will apply to all selected tests",
                                                "Copy the reported failure information to the selected tests"]
                                   )
        self.allKnownBugFiles = []
        self.rootSuites = []

    def _getStockId(self):
        return "find"

    def _getTitle(self):
        return "Find Failure Information"

    def getDialogTitle(self):
        return "Find, copy and move information for automatic interpretation of test failures"

    def findAllBugs(self, bugMap):
        test = self.currTestSelection[0]
        bugs = CheckForBugs().findAllBugs(test, test.stateInGui, bugMap)[0]
        return [str(b.bugInfo) for b in bugs]

    def updateOptions(self):
        # Only do this on completed tests
        if not all((test.stateInGui.isComplete() for test in self.currTestSelection)):
            return False

        self.notify("ActionStart")
        bugMap = BugMap()
        self.rootSuites = self.findSelectedRootSuites()
        self.allKnownBugFiles = self.findAllKnownBugsFiles()
        for bugFile in self.allKnownBugFiles:
            self.notify("ActionProgress")
            bugMap.readFromFile(bugFile)

        # We assume the first test is representative and only check all the bugs on that one
        bugs = self.findAllBugs(bugMap)
        if bugs:
            self.optionGroup.setPossibleValues("bug", bugs)
            self.optionGroup.setValue("bug", bugs[0])
        self.notify("ActionStop")
        return False

    def findAllKnownBugsFiles(self):
        files = []

        def progress():
            self.notify("ActionProgress")
        for app in self.currAppSelection:
            for fileName in app.getFileNamesFromFileStructure("knownbugs", progress):
                if fileName not in files:
                    files.append(fileName)
        return files

    def findSelectedRootSuites(self):
        roots = []
        for test in self.currTestSelection:
            root = test.getAllTestsToRoot()[0]
            if root not in roots:
                roots.append(root)
        return roots

    def findBugFileTest(self, filePath):
        testDir = os.path.dirname(filePath)
        for suite in self.rootSuites:
            if testDir.startswith(suite.getDirectory()):
                relPath = plugins.relpath(testDir, suite.getDirectory())
                return suite.findSubtestWithPath(relPath)

    def getTestsToApplyTo(self, copyChoice, bugFile):
        if copyChoice == 0:
            return self.rootSuites
        elif copyChoice == 1:
            bugFileTest = self.findBugFileTest(bugFile)
            return ReportBugs.findCommonSelectedAncestors(self.currTestSelection + [bugFileTest])
        else:
            return self.currTestSelection

    def getFileNames(self, suitesOrTests, bugFile):
        fileNames = []
        for suiteOrTest in suitesOrTests:
            name = os.path.basename(bugFile)
            fileName = os.path.join(suiteOrTest.getDirectory(), name)
            if not any((fileName.startswith(f) for f in fileNames)):
                fileNames.append(fileName)

        return fileNames

    def findBugInfo(self, bugStr):
        for bugFile in self.allKnownBugFiles:
            parser = BugMap.makeParser(bugFile)
            for section in parser.sections():
                if (parser.has_option(section, "bug_id") and parser.get(section, "bug_id") == bugStr) or \
                   (parser.has_option(section, "brief_description") and parser.get(section, "brief_description") == bugStr):
                    newParser = ConfigParser()
                    newParser.add_section(section)
                    for key, value in parser.items(section):
                        newParser.set(section, key, value)
                    bugMap = BugMap()
                    bugMap.readFromParser(newParser)
                    if bugStr in self.findAllBugs(bugMap):
                        return bugFile, section, parser, newParser, bugMap

    def performOnCurrent(self):
        bugFile, section, parser, newParser, bugMap = self.findBugInfo(self.optionGroup.getValue("bug"))
        copyChoice = self.optionGroup.getValue("copy")
        suitesOrTests = self.getTestsToApplyTo(copyChoice, bugFile)
        newFileNames = self.getFileNames(suitesOrTests, bugFile)
        if copyChoice != 2:
            parser.remove_section(section)
            if len(parser.sections()) == 0:
                if len(newFileNames) == 1 and not os.path.isfile(newFileNames[0]):
                    self.movePath(bugFile, newFileNames[0])
                    newFileNames = []  # Don't need to write parser also
                else:
                    self.removePath(bugFile)
            else:
                parser.write(open(bugFile, "w"))

        for fileName in newFileNames:
            with open(fileName, "a") as f:
                newParser.write(f)

        applyBugsToTests(self.currTestSelection, bugMap)

    @staticmethod
    def removePath(dir):
        return plugins.removePath(dir)

    @staticmethod
    def movePath(oldPath, newPath):
        # overridden by version control modules
        os.rename(oldPath, newPath)


def getInteractiveActionClasses():
    return [ApproveTests, KillTests, MarkTest, UnmarkTest, LoadFromRerun, RecomputeTests, ReportBugsAndRecompute,
            SuspendTests, UnsuspendTests, SplitResultFiles, FindKnownBugs]
